// Copyright 2009-2010 Sandia Corporation. Under the terms
// of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
// Government retains certain rights in this software.
// 
// Copyright (c) 2009-2010, Sandia Corporation
// All rights reserved.
// 
// This file is part of the SST software package. For license
// information, see the LICENSE file in the top level directory of the
// distribution.


#include <PHXSimC.h>
#include "sst/core/serialization/element.h"
#include <boost/serialization/deque.hpp>

#if 0
#define DBG( fmt, args... ) \
    m_dbg.write( "%s():%d: "fmt, __FUNCTION__, __LINE__, ##args)
#else
#define DBG( fmt, args... ) ;
#endif

#define ERR( fmt, args...) \
  printf("%s():%d: "fmt, __FUNCTION__, __LINE__, ##args);

PHXSimC::PHXSimC( ComponentId_t id, Params_t& params ) :
    Component( id ),
    m_printStats("no"),
    m_dbg( *new Log< PHXSIMC_DBG >( "PHXSimC::", false ) ),
    m_log( *new Log< >( "INFO PHXSimC: ", false ) )
{
    std::string frequency = "2.2 GHz";

    engines = 0;
    if ( params.find( "engines" ) != params.end() ) {
      sscanf(params["engines"].c_str(), "%d", &engines);
    }
    
    linksToUse = 8;
    if ( params.find( "linksToUse" ) != params.end() ) {
      sscanf(params["linksToUse"].c_str(), "%d", &linksToUse);
    }
    if (linksToUse < 1 || linksToUse > 8) {
      printf("%d is invalid number of links for PHX", linksToUse);
      exit(-1);
    }

    if ( params.find( "info" ) != params.end() ) {
        if ( params[ "info" ].compare( "yes" ) == 0 ) {
            m_log.enable();
        }
    }

    if ( params.find( "debug" ) != params.end() ) {
        if ( params[ "debug" ].compare( "yes" ) == 0 ) {
            m_dbg.enable();
        }
    }

    if ( params.find( "clock" ) != params.end() ) {
      frequency = params["clock"];
    }

    DBG("new id=%lu\n",id);

    m_memChan = new memChan_t( *this, params, "bus" );

    // connect adv mem link
    advMem = configureLink("advBus0", "1 Ghz");

    // initialize queues
    if (engines > 0 && (advMem==0)) {
      printf("Engines Requested, but no Advanced Memory Link for PHXMem");
    } 
    if (engines > 0) {
      deque<PHXEvent*> q(0);
      for (int i = 0; i < engines; ++i) {
	advMemEvents.push_back(q);
	engOutstandingReads.push_back(0);
	engIdx.push_back(0);
      }
    }

    transaction_id = 0;
    hadFirstEvent = 0;
    resetCounters();

    /*ClockHandler_t* handler;
    handler = new EventHandler< PHXSimC, bool, Cycle_t >
    ( this, &PHXSimC::clock );*/

    m_log.write("freq %s\n",frequency.c_str());
    //TimeConverter* tc = registerClock( frequency, handler );
    TimeConverter* tc = registerClock( frequency, new Clock::Handler<PHXSimC>(this, &PHXSimC::clock) );
    m_log.write("period %ld\n",tc->getFactor());


    m_memorySystem = new PHXSim::SingleCube();
	  //getSingleCubeInstance();
    if ( ! m_memorySystem ) {
      _abort(PHXSimC,"SingleCube() failed\n");
    }
//	 PHXSim::Callback< PHXSimC, void, uint64_t,uint64_t >* readDataCB;
    PHXSim::Callback< PHXSimC, void, PHXSim::Packet *,uint64_t >* packetReturnCB;

    packetReturnCB = new PHXSim::Callback< PHXSimC, void, PHXSim::Packet *,uint64_t >
      (this, &PHXSimC::packetReturn);
    m_memorySystem->RegisterPacketCallback( packetReturnCB );
}

int PHXSimC::Finish() 
{
    if ( ! m_printStats.compare("yes" ) ) {
//        m_memorySystem->printStats();
    }

    //print Counters
    printf("Engine: engFPMult %llu\n", engFPMul);
    printf("Engine: engFPDiv %llu\n", engFPDiv);
    printf("Engine: engFPAdd %llu\n", engFPAdd);
    // add in the 'partial' reads (sub-cacheline)
    engRW += (engPartialR + 63) / 64;
    printf("Engine: engRW %llu\n", engRW);
    printf("Engine: engDidWork %llu\n", engDidWork);
    printf("Engine: engFinishedUnit %llu\n", engFinishedUnit);

    return 0;
}

void PHXSimC::packetReturn(PHXSim::Packet *pkt, uint64_t clockcycle)
{
    // ACK any write tags that came in at the tail of this packet
	DBG("CALLBACK type %d\n",pkt->packetType);
    std::map<unsigned, PHXSim::Packet *>::iterator it; 
    transToEngMap_t::iterator ei;
    for (size_t i=0; i<pkt->returnTags.size(); i++) 
    {
      if ( (it = writeReturnMap.find(pkt->returnTags[i])) 
	   != writeReturnMap.end()) {
	// regular write ack
        memChan_t::event_t* event = new memChan_t::event_t;
	
	if ((*it).second->address == 0) 
	  printf("warning, null write return address\n");
        event->addr = (*it).second->address;
        event->reqType = memChan_t::event_t::WRITE; 
        event->msgType = memChan_t::event_t::RESPONSE; 
        if( ! m_memChan->send(event) ) {
	  _abort(PHXSimC::packetReturn, "m_memChan->send failed\n");
        }
      } else if ((ei = writeTransToEng.find(pkt->returnTags[i]))
		 != writeTransToEng.end()) {
	// engine write
	//printf("got back eng write ack\n");
	writeTransToEng.erase(ei);
      } else {
	// fail
	_abort(PHXSimC::packetReturn,"got back tag for unissued packet?\n");
      }
    }
    if (pkt->returnTags.size() > 0) {
        DBG("Returned %d return tags from packet\n", pkt->returnTags.size());
    }

    // check if this is a engine packet return
    if (pkt->packetType == PHXSim::RETURN_DATA) {
      transToEngMap_t::iterator mi = readTransToEng.find(pkt->packetID);
      if (mi != readTransToEng.end()) {	
	//printf("Found %d from Eng %d\n", mi->first, mi->second);
	// decrement outstanding reads
	engOutstandingReads[mi->second]--;
	// clean up
	readTransToEng.erase(mi);
	delete pkt;
	return;
      }
    }

    // now handle the main contents of this packet 
    memChan_t::event_t* other_event = new memChan_t::event_t;

    other_event->addr = pkt->address;
    switch (pkt->packetType) 
    {
        case PHXSim::RETURN_DATA:
	    other_event->reqType = memChan_t::event_t::READ; 
            other_event->msgType = memChan_t::event_t::RESPONSE; 
            DBG("READ DATA: for read at addr %x\n", other_event->addr);
            /* SingleCube also queues up packets into the link read return
             * buffer, which we should clear out to save memory */

            // FIXME: iterators here? 
            for (size_t i=0; i< m_memorySystem->readReturnBuffer.size(); i++)
            {
                if (m_memorySystem->readReturnBuffer[i].packetID == pkt->packetID) {
                    m_memorySystem->readReturnBuffer.erase(m_memorySystem->readReturnBuffer.begin() + i);
                    break;
                }
            }
            break;

            //for now these aren't issued by the simulator
        case PHXSim::TAG_RETURN:
            if ( (it = writeReturnMap.find(pkt->packetID)) == writeReturnMap.end()) {
                _abort(PHXSimC::packetReturn, "got back tag for unissued packet?\n");
            }

            other_event->addr = (*it).second->address;
            other_event->reqType = memChan_t::event_t::WRITE; 
            other_event->msgType = memChan_t::event_t::RESPONSE; 

            DBG("WRITE ACK: for write at addr %x\n", other_event->addr);
            break;
        case PHXSim::TOKEN_RETURN:
            //throw these away
            DBG("Garbage TOKEN RETURN\n");
            delete(other_event);
	    delete pkt;
            return;
            break;

        default:
            delete(other_event);
            _abort(PHXSimC::packetReturn, "Shouldn't have gotten back this type of packet here %d %d\n", pkt->packetType, PHXSim::TOKEN_RETURN);
    }
    if( ! m_memChan->send(other_event) ) {
        _abort(PHXSimC::packetReturn, "m_memChan->send failed\n");
    }
    DBG("Sent all events back successfully\n");

    delete pkt;
}

// add a write from an engine
void PHXSimC::engWrite(unsigned int addr, int eng, int size) {
  if (size == 64) {
    engRW++;
  } else {
    engPartialR += size;
    if (size < 8) size = 8; // reads <8 don't seem to work
  }
  //Create a packet
   PHXSim::Packet *outgoingPacket = 
    new PHXSim::Packet( PHXSim::DATA_WRITE, size, addr, transaction_id);
   // record it
   writeTransToEng[transaction_id] = eng;
   transaction_id++;
   // add to the transaction Q
   m_transQ.push_back( outgoingPacket );
   //printf("issued Eng Write %d\n", transaction_id);
}

// add a read from an engine
void PHXSimC::engRead(unsigned int addr, int eng, int size) {
  if (size == 64) {
    engRW++;
  } else {
    engPartialR += size;
    if (size < 8) size = 8; // reads <8 don't seem to work
  }
  // Create a packet
  PHXSim::Packet *outgoingPacket = 
    new PHXSim::Packet( PHXSim::DATA_READ, size, addr, transaction_id);
  // record it
  readTransToEng[transaction_id] = eng;
  //printf("Issued %d from eng %d\n", transaction_id, eng);
  transaction_id++;
  // add to the transaction Q
  m_transQ.push_back( outgoingPacket );

  // increment the engine's couter
  engOutstandingReads[eng]++;
}

// bundle together what we can
static void thin(vector<unsigned int> &list) {
  sort(list.begin(), list.begin());
  set<unsigned int> addrsToSend;
  for (unsigned int i = 0; i < list.size(); ++i) {
    unsigned int addr = list[i];
    addr = (addr >> 6) << 6;
    addrsToSend.insert(addr);
  }
  if (addrsToSend.size() < list.size()) {
    //printf("thined from %ld to %ld\n", list.size(),addrsToSend.size());
    list.clear();
    for (set<unsigned int>::iterator si = addrsToSend.begin();
	 si != addrsToSend.end(); ++si) {
      list.push_back(*si);
    }
  }
}

bool PHXSimC::handleForceCalc(PHXEvent *pe, int eng) {
  switch (pe->state) {
  case INIT:
    engOutstandingReads[eng] = 0; // track how many we have oustanding
    engIdx[eng] = 0; // used to hold countdown for FP latency
    // thin the address lists
    thin(pe->fAddrs.inits);
    thin(pe->fAddrs.a_fs);
    pe->state = ISSUE_INITS;
    break;

  case ISSUE_INITS:
    {
      // issue the initial requests ([xyz]tmp - a.x[j][0-2])
      for (unsigned int i = 0; i < pe->fAddrs.inits.size(); ++i) {
	engRead(pe->fAddrs.inits[i], eng, 32);  //32?
      }
      // calculate FP delay once we get things back
      // (3 subtractions, then 3 multiplies, then two adds) all pipelined
#if 0
      engIdx[eng] = (3 + FPAddLat - 1) +
	(3 + FPMultLat - 1) +
	(2 + FPAddLat - 1);
#else
      engIdx[eng] = (8 + FPMultLat - 1);
#endif
      engFPMul +=3;
      engFPAdd += 5;
      // advance state
      pe->state = WAIT_INITS_AND_CALC;
      break;
    }
  case WAIT_INITS_AND_CALC:
    {
      // check for completion
      if (engOutstandingReads[eng] == 0) {
	engIdx[eng]--; // wait for FP to finish
	if (engIdx[eng] == 0) {
	  // advance state
	  pe->state = CHECK_BRANCH;
	  break;
	}
      }
      break;
    }
  case CHECK_BRANCH:
    {
      // checking the if (rsq < cutforcesq)  branch
      if (pe->tookBranch) {
	// continue with the algorithm
	// calculate FP latency
	// one divide; then 2 multiplies; then one subtract; then 2 mult
#if 0
	engIdx[eng] = (1 + FPDivLat - 1) +
	  (2 + FPMultLat - 1) +
	  (1 + FPAddLat - 1) +
	  (2 + FPMultLat - 1);
#else
	engIdx[eng] = (FPDivLat) +
	  (5 + FPMultLat - 1);
#endif
	engFPDiv++;
	engFPMul += 4;
	engFPAdd++;
	pe->state = CALC_FORCE;
	break;
      } else {
	// we're done!
	return true;
      }
    }
  case CALC_FORCE:
    engIdx[eng]--; // wait for FP to finish
    if (engIdx[eng] == 0) {
      pe->state = ISSUE_AF_READS;
    } else {
      break;
    }
  case ISSUE_AF_READS:
    {
      // issue the force reads
      for (unsigned int i = 0; i < pe->fAddrs.a_fs.size(); ++i) {
	engRead(pe->fAddrs.a_fs[i], eng);
      }
      // calculate FP latency
      // 3 multiplies, 6 add/sub
#if 0
      engIdx[eng] = (3 + FPMultLat - 1) +
	(6 + FPAddLat - 1);
#else
      engIdx[eng] = (9 + FPMultLat - 1);
#endif
      engFPMul += 3;
      engFPAdd += 6;
      pe->state = WAIT_AF_READS_AND_CALC;
    }
    break;
  case WAIT_AF_READS_AND_CALC:
    // check for completion
    if (engOutstandingReads[eng] == 0) {
      engIdx[eng]--; // wait for FP to finish
      if (engIdx[eng] == 0) {
	// advance state
	pe->state = ISSUE_SENDS;
	break;
      }
    }
    break;
  case ISSUE_SENDS:
    {
      // issue the force reads
      for (unsigned int i = 0; i < pe->fAddrs.a_fs.size(); ++i) {
	engWrite(pe->fAddrs.a_fs[i], eng);
      }
      return true;
      break;
    }

  case LAST_STATE:
  default:
    ERR("Bad state");
  }

  return false;
}

// return if that event is done and should be disposed of
// Note: we should probably throttle how many reads are sent at a time
// and overlap the I and X arrays. 
bool PHXSimC::handleMatVec(PHXEvent *pe, int eng) {
  switch (pe->state) {
  case INIT:
    engOutstandingReads[eng] = 0; // track how many we have oustanding
    engIdx[eng] = 0; // used to hold countdown for FP latency
    // sort the X list, just to make sure
    sort(pe->mvAddrs.X.begin(), pe->mvAddrs.X.begin());
  case ISSUE_I_AND_CV:
    {
      const int FPPerRead = 64/8;
      const int intPerRead = 64/4;
      // figure out how many I and CV refs to make
      int neededIs = (pe->mvAddrs.I.size() + intPerRead - 1) / intPerRead;
      int neededCVs = (pe->mvAddrs.CV.size() + FPPerRead - 1) / FPPerRead;      
      // make read requests
      // I requests
      for (int i = 0; i < neededIs; ++i) {
	unsigned int addr = pe->mvAddrs.I[i * intPerRead];
	engRead(addr, eng);	
      }
      // CV requests
      for (int i = 0; i < neededCVs; ++i) {
	unsigned int addr = pe->mvAddrs.CV[i * FPPerRead];
	engRead(addr, eng);
      }

      // advance state
      pe->state = WAIT_I_AND_CV;
    }
    break;
  case WAIT_I_AND_CV:
    {
      //wait till we have 'em all
      if (engOutstandingReads[eng] == 0) {
	pe->state = ISSUE_X;
      }
    }
    break;
  case ISSUE_X:
    {
      set<unsigned int> addrsToSend;
      // bundle together what we can
      for (unsigned int i = 0; i < pe->mvAddrs.X.size(); ++i) {
	unsigned int addr = pe->mvAddrs.X[i];
	addr = (addr >> 6) << 6;
	addrsToSend.insert(addr);
      }

      // send requests
      for (set<unsigned int>::iterator i = addrsToSend.begin();
	   i != addrsToSend.end(); ++i) {
	engRead(*i,eng);
      }
      // compute FP latency
      engIdx[eng] = pe->mvAddrs.X.size() + FPMultLat - 1;
      engFPMul += pe->mvAddrs.X.size();
      // advance state
      pe->state = WAIT_X;
    }
    break;
  case WAIT_X:
    {
      // check for completion
      if (engOutstandingReads[eng] == 0) {
	engIdx[eng]--; // wait for FP to finish
	if (engIdx[eng] == 0) {
	  return true;
	}
      }
    }
    break;

  case LAST_STATE:
  default:
    ERR("Bad state");
  }

  return false;
}

bool PHXSimC::handlePagerank(PHXEvent *pe, int eng) {
  switch (pe->state) {
  case INIT:
    engOutstandingReads[eng] = 0; // track how many we have oustanding
    engIdx[eng] = 0; // used to hold countdown for FP latency
    thin(pe->pAddrs.neigh);
    pe->pAddrs.num_care_neigh = pe->pAddrs.care_neigh.size();
    thin(pe->pAddrs.care_neigh);
    pe->state = ISSUE_INITS;
    break;
  case ISSUE_INITS:
    {
      for (unsigned int i = 0; i < pe->pAddrs.inits.size(); ++i) {
	engRead(pe->pAddrs.inits[i], eng, 4);
      }
      pe->state = CHECK_BRANCH;
      break;
    }
  case CHECK_BRANCH:
    {
      // check for completion
      if (engOutstandingReads[eng] == 0) {
	// check branch
	if (pe->tookBranch) {
	  pe->state = ISSUE_NEIGH;
	  break;
	} else {
	  // we're done
	  return true;
	}
      } else {
	break;
      }
    }
  case ISSUE_NEIGH:
    {
      for (unsigned int i = 0; i < pe->pAddrs.neigh.size(); ++i) {
	engRead(pe->pAddrs.neigh[i], eng, 32);  
      }
      pe->state = WAIT_NEIGH;
      break;
    }
  case WAIT_NEIGH:
    {
       if (engOutstandingReads[eng] == 0) {
	 pe->state = ISSUE_CARE_NEIGH;
       }
       break;
    }
  case ISSUE_CARE_NEIGH:
    {
      for (unsigned int i = 0; i < pe->pAddrs.care_neigh.size(); ++i) {
	engRead(pe->pAddrs.care_neigh[i], eng, 32);  
      }
      // calculate FP latency
      // one divide and one add per neighbor
      int numN = pe->pAddrs.num_care_neigh;
      engIdx[eng] = (FPDivLat + numN - 1) + (FPAddLat + numN - 1);
      engFPDiv += numN;
      engFPAdd += numN;
      // continue
      pe->state = WAIT_CARE_NEIGH_AND_CALC;      
      break;
    }
  case WAIT_CARE_NEIGH_AND_CALC:
    {
      if (engOutstandingReads[eng] == 0) {
	engIdx[eng]--; // wait for FP to finish
	if (engIdx[eng] == 0) {
	  // advance state
	  pe->state = ISSUE_SEND;
	}
      }
      break;
    }
  case ISSUE_SEND:
    engWrite(pe->pAddrs.total, eng, 8);
    return true;
    break;
  case LAST_STATE:
  default:
    ERR("Bad state");
  }

  return false;
}

void PHXSimC::advanceEngine(int eng) {
  PHXEvent *pe = advMemEvents[eng].front();
 
  bool done = 0;
  switch (pe->eventType) {
  case MATVEC:
    done = handleMatVec(pe, eng);
    break;
  case FORCE_CALC:
    done = handleForceCalc(pe, eng);
    break;
  case PAGERANK:
    done = handlePagerank(pe, eng);
    break;
  case LAST_EVENT:
  default:
    printf("Unknown event type in " __FILE__ ":%d\n", __LINE__);
    exit(-1);
  }

  if (done) {
    //printf("finishing event\n");
    engFinishedUnit++;
    // clean up the queue
    advMemEvents[eng].pop_front();
    // send back the event
    advMem->Send(1,pe);
  }
}

void PHXSimC::resetCounters() {
   engFPMul = 0;
   engFPDiv = 0;
   engFPAdd = 0;
   engRW = 0;
   engPartialR = 0;
   engDidWork = 0;
   engFinishedUnit = 0;
   printf("PHX Reset Counters\n");
}

void PHXSimC::handleAdvMem() {
  if (advMem == 0) return;

  // check for any incoming advanced memory operations
  Event *e = 0;
  do {
    Event *e = advMem->Recv();
    if (e) {
      hadFirstEvent = 1;
      PHXEvent *pe = dynamic_cast<PHXEvent*>(e);
      if (pe) {
	//printf("PHXSim got adv memory event\n");
	if (pe->eventType == RESET) {
	  resetCounters();
	  delete e;
	} else {
	  // add it to an engine's queue
	  pe->state = INIT;
	  advMemEvents[pe->eventID % engines].push_back(pe);
	}
      } else {
	printf("PHXSimC Advanced Memory Link got back wrong type of event\n");
	exit(-1);
      }
    }
  } while (e);

  // advance pending events
  for (int eng = 0; eng < engines; ++eng) {
    if (!advMemEvents[eng].empty()) {
      engDidWork++;
      advanceEngine(eng);
    }
  }
}

bool PHXSimC::clock( Cycle_t current )
{
	static unsigned link_id = 0;
	PHXSim::Packet *outgoingPacket;

	/* A memory system update is actually a link update, which happens at 4x the
	 * rate of the cube clock.
	 *
	 * Additionally, sending packets to a cube must be aligned to the cube clock
	 * -- hence why update should be called 4x here before adding packets. 
	 * 
	 * The SDL file should set the clock of this component to 1.0GHz in order to
	 * get the 4GHz link clock
	 */
	// don't do anything until we get an event (gives us a cheap
	// 'fast forward')
	if (hadFirstEvent != 0) {
	  m_memorySystem->Update();
	  m_memorySystem->Update();
	  m_memorySystem->Update();
	  m_memorySystem->Update();
	}

    // handle new and outstanding adv mem ops
    handleAdvMem();

    // check for incoming events from processor
    memChan_t::event_t* event;
    while ( m_memChan->recv( &event ) ) 
    {
      hadFirstEvent = 1;
        DBG("got an event %x\n", event);

        PHXSim::PacketType pktType = convertType( event->reqType );
        DBG("transType=%d addr=%#lx tag=%d\n", pktType, event->addr, transaction_id );
        outgoingPacket = new PHXSim::Packet( pktType, 64, event->addr, transaction_id);
        m_transQ.push_back( outgoingPacket );

        // store write packets, we will need them later when the tags come back
        if ( pktType == PHXSim::DATA_WRITE )
        {
            if (writeReturnMap.find(transaction_id) != writeReturnMap.end()) {
                _abort(PHXSimC,"Write tag being re-added, something is wrong\n");
            }
            writeReturnMap[transaction_id] = outgoingPacket;
        }
        transaction_id++;
        delete event;
    }

    int ret = 1; 
    while ( ! m_transQ.empty() && ret ) {
		 PHXSim::Packet *pkt = m_transQ.front();
        if (  ( ret = m_memorySystem->AddPacket( pkt, link_id ) ) )
        {
	        DBG("AddPacket succeeded %#x to link %d\n", pkt->address, link_id);
	        m_transQ.pop_front();
		//FIXME: Need to make this mod NUM_LINKS, but that's broken at the moment
			  link_id = (link_id+1) % linksToUse; // PHXSim::NUM_LINKS;
        } else {
	        DBG("AddPacket failed\n");
        }
    }
    return false;
}

extern "C" {
PHXSimC* PHXSimCAllocComponent( SST::ComponentId_t id,
                                    SST::Component::Params_t& params )
{
    return new PHXSimC( id, params );
}
}

BOOST_CLASS_EXPORT(PHXEvent)

#if WANT_CHECKPOINT_SUPPORT
BOOST_CLASS_EXPORT(PHXSimC)

BOOST_CLASS_EXPORT_TEMPLATE4( SST::EventHandler,
                                PHXSimC, bool, SST::Cycle_t, SST::Time_t )
#endif
