// Copyright 2009-2010 Sandia Corporation. Under the terms
// of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
// Government retains certain rights in this software.
// 
// Copyright (c) 2009-2010, Sandia Corporation
// All rights reserved.
// 
// This file is part of the SST software package. For license
// information, see the LICENSE file in the top level directory of the
// distribution.


#ifndef MPI_H_
#define MPI_H_

#if defined(c_plusplus) || defined(__cplusplus)
extern "C" {
#endif

typedef struct _status   MPI_Status;
typedef int              MPI_Comm;
typedef int              MPI_Datatype;
typedef struct _request *MPI_Request;
typedef long		 MPI_Aint;
typedef int		 MPI_Fint;
typedef int              MPI_Errhandler;

typedef enum {
  MPI_MAX=26,
  MPI_MIN=27,
  MPI_SUM=28,
  MPI_PROD=29,
  MPI_LAND=30,
  MPI_BAND=31,
  MPI_LOR=32,
  MPI_BOR=33,
  MPI_LXOR=34,
  MPI_BXOR=35,
  MPI_MAXLOC=36,
  MPI_MINLOC=37,
  MPI_REPLACE=38
} MPI_Op;

/*
 * misc. constants
 */
#define MPI_SUCCESS		0		/* no errors */
#define MPI_ANY_SOURCE		-1		/* match any source rank */
#define MPI_CANCEL_SOURCE	-3		/* successful cancel */
#define MPI_ANY_TAG		-1		/* match any message tag */
#define MPI_BOTTOM		((void *) 0)	/* base reference address */
#define MPI_UNDEFINED           (-1)

/*
 * error classes
 */
#define MPI_ERR_BUFFER		1	/* invalid buffer pointer */
#define MPI_ERR_COUNT		2	/* invalid count argument */
#define MPI_ERR_TYPE		3	/* invalid datatype argument */
#define MPI_ERR_TAG		4	/* invalid tag argument */
#define MPI_ERR_COMM		5	/* invalid communicator */
#define MPI_ERR_RANK		6	/* invalid rank */
#define MPI_ERR_REQUEST		7	/* invalid request handle */
#define MPI_ERR_ROOT		8	/* invalid root */
#define MPI_ERR_GROUP		9	/* invalid group */
#define MPI_ERR_OP		10	/* invalid operation */
#define MPI_ERR_TOPOLOGY	11	/* invalid topology */
#define MPI_ERR_DIMS		12	/* invalid dimension argument */
#define MPI_ERR_ARG		13	/* invalid argument */
#define MPI_ERR_UNKNOWN		14	/* unknown error */
#define MPI_ERR_TRUNCATE	15	/* message truncated on receive */
#define MPI_ERR_OTHER		16	/* LAM error */
#define MPI_ERR_INTERN		17	/* internal MPI error */
#define MPI_ERR_IN_STATUS	18	/* error code is in status */
#define MPI_ERR_PENDING		19	/* pending request */
#define MPI_ERR_SYSRESOURCE	20	/* out of system resources */
#define MPI_ERR_LOCALDEAD	21	/* process in local group is dead */
#define MPI_ERR_REMOTEDEAD	22	/* process in remote group is dead */
#define MPI_ERR_VALUE		23	/* truncated info value */
#define MPI_ERR_FLAGS		24	/* mismatched run-time flags */
#define MPI_ERR_SERVICE		25	/* name publishing service error */
#define MPI_ERR_NAME		26	/* name not published */
#define MPI_ERR_SPAWN		27	/* error while spawning processes */
#define MPI_ERR_KEYVAL		28	/* invalid key value */
#define MPI_ERR_INFO_NOKEY	29	/* no such info key */
#define MPI_ERR_WIN		30	/* invalid window */
#define MPI_ERR_EPOCH		31	/* mismatched one-sided synch. */
#define MPI_ERR_TYPENOTSUP	32	/* operation not supported on type */
#define MPI_ERR_INFO_KEY	33	/* invalid info key */
#define MPI_ERR_INFO_VALUE	34	/* invalid info value */
#define MPI_ERR_NO_MEM		35	/* no memory left */
#define MPI_ERR_BASE		36	/* invalid base pointer value */
#define MPI_ERR_LASTCODE	37	/* last error code */

/*
 * NULL handles
 */
#define MPI_COMM_NULL           ((MPI_Comm) 0)
#define MPI_DATATYPE_NULL       ((MPI_Datatype) 0)
#define MPI_REQUEST_NULL        ((MPI_Request) 0)
#define MPI_OP_NULL             ((MPI_Op) 0)

#define MPI_STATUSES_IGNORE     ((MPI_Status*) NULL)
#define MPI_STATUS_IGNORE       ((MPI_Status*) NULL)

/*
 * MPI_Init_thread constants
 */
#define MPI_THREAD_SINGLE		0
#define MPI_THREAD_FUNNELED		1
#define MPI_THREAD_SERIALIZED		2
#define MPI_THREAD_MULTIPLE		3

/*
 * Transparent MPI data type.
 */
struct _status { 
	int		MPI_SOURCE;
	int		MPI_TAG;
	int		MPI_ERROR;
	int		st_length;		/* message length */
};

/*
 * MPI predefined handles
 */
#define MPI_COMM_WORLD          1
#define MPI_COMM_SELF		0

#define MPI_BYTE   0
#define MPI_CHAR   1
#define MPI_SHORT  2
#define MPI_INT    3
#define MPI_LONG   4
#define MPI_FLOAT  5
#define MPI_DOUBLE 6
#define MPI_TYPE_MAX 7

extern int MPI_Abort(MPI_Comm, int);
extern int MPI_Allreduce(void *, void *, int, MPI_Datatype, 
                         MPI_Op, MPI_Comm);
extern int MPI_Barrier(MPI_Comm);
extern int MPI_Bcast(void *, int, MPI_Datatype, int, MPI_Comm);
extern int MPI_Comm_rank(MPI_Comm, int *);
extern int MPI_Comm_size(MPI_Comm, int *);
extern int MPI_Finalize(void);
extern int MPI_Finalized(int *flag);
extern int MPI_Get_count(MPI_Status *, MPI_Datatype, int *);
extern int MPI_Get_elements(MPI_Status *, MPI_Datatype, int *);
extern int MPI_Init(int *, char ***);
extern int MPI_Initialized(int *);
extern int MPI_Init_thread(int *, char ***, int, int *);
extern int MPI_Iprobe(int, int, MPI_Comm, int *, MPI_Status *);
extern int MPI_Irecv(void *, int, MPI_Datatype, int, int, MPI_Comm, 
                     MPI_Request *);
extern int MPI_Irsend(void *, int, MPI_Datatype, int, int, MPI_Comm, 
                      MPI_Request *);
extern int MPI_Isend(void *, int, MPI_Datatype, int, int, MPI_Comm, 
                     MPI_Request *);
extern int MPI_Issend(void *, int, MPI_Datatype, int, int, MPI_Comm, 
                      MPI_Request *);
extern int MPI_Probe(int, int, MPI_Comm, MPI_Status *);
extern int MPI_Recv(void *, int, MPI_Datatype, int, int, MPI_Comm, 
                    MPI_Status *);
extern int MPI_Reduce(void *, void *, int, MPI_Datatype, MPI_Op, int, 
                      MPI_Comm);
extern int MPI_Rsend(void *, int, MPI_Datatype, int, int, MPI_Comm);
extern int MPI_Send(void *, int, MPI_Datatype, int, int, MPI_Comm);
extern int MPI_Ssend(void *, int, MPI_Datatype, int, int, MPI_Comm);
extern int MPI_Testall(int, MPI_Request *, int *, MPI_Status *);
extern int MPI_Testany(int, MPI_Request *, int *, int *, 
                       MPI_Status *);
extern int MPI_Test(MPI_Request *, int *, MPI_Status *);
extern int MPI_Testsome(int, MPI_Request *, int *, int *, 
                        MPI_Status *);
extern int MPI_Waitall(int, MPI_Request *, MPI_Status *);
extern int MPI_Waitany(int, MPI_Request *, int *, MPI_Status *);
extern int MPI_Wait(MPI_Request *, MPI_Status *);
extern int MPI_Waitsome(int, MPI_Request *, int *, int *, 
                        MPI_Status *);
extern double MPI_Wtick(void);
extern double MPI_Wtime(void);

#if defined(c_plusplus) || defined(__cplusplus)
}
#endif

#endif	/* _MPI_H */
