// Copyright 2009-2010 Sandia Corporation. Under the terms
// of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
// Government retains certain rights in this software.
// 
// Copyright (c) 2009-2010, Sandia Corporation
// All rights reserved.
// 
// This file is part of the SST software package. For license
// information, see the LICENSE file in the top level directory of the
// distribution.


#include "mpi_internal.h"

#include <stdlib.h>
#include <string.h>

#define BARRIER_TAG   (-10)
#define BCAST_TAG     (-11)
#define REDUCE_TAG    (-12)


/* reduction operation code */
template <class T>
int doOp(MPI_Op op, T *inbuf, T *rbuf, int count) {
  int i;
  T *a = (T*)inbuf;
  T *b = (T*)rbuf;

#define FUNC_FUNC(f)				\
  for (i = 0; i < count; ++i) {			\
    *(b) = current_func(*(b),*(a));		\
  }						\
  break;

#define OP_FUNC(o)				\
  for (i = 0; i < count; ++i) {			\
    *(b) o *(a);				\
  }						\
  break;

#define OP_FUNC2(o)				\
  for (i = 0; i < count; ++i) {			\
    int _t = (int)*(b);				\
    (_t) o ((int)*(a));				\
    *b = _t;					\
  }						\
  break;

  switch (op) {
  case  MPI_MAX:
#undef current_func
#define current_func(a, b) ((a) > (b) ? (a) : (b))    
    FUNC_FUNC(max);

  case  MPI_MIN: 
#undef current_func
#define current_func(a, b) ((a) < (b) ? (a) : (b))    
    FUNC_FUNC(min);

  case  MPI_SUM: OP_FUNC(+=);

  case  MPI_PROD: OP_FUNC(*=);

  case  MPI_LAND:  
#undef current_func
#define current_func(a, b) ((a) && (b))
    FUNC_FUNC(land);

  case  MPI_BAND: OP_FUNC2(&=);

  case  MPI_LOR:
#undef current_func
#define current_func(a, b) ((a) || (b))
    FUNC_FUNC(lor);

  case  MPI_BOR: OP_FUNC2(|=);

  case  MPI_LXOR:
#undef current_func
#define current_func(a, b) ((a ? 1 : 0) ^ (b ? 1: 0))
    FUNC_FUNC(lxor);

  case  MPI_BXOR: OP_FUNC2(^=);

  case  MPI_MAXLOC:
  case  MPI_MINLOC:
  case  MPI_REPLACE:
  default:
    printf("unknown/unimplemented operator %d\n", op);
    return MPI_ERR_OP;
  }

  return MPI_SUCCESS;
}


static inline int
mpi_int_reduce(void *sbuf, void *rbuf, int count, MPI_Datatype dtype, 
               MPI_Op op, int root, MPI_Comm comm)
{
    int i, size, rank, ret;
    char *buffer = NULL;
    char *inbuf;
    MPI_Request req;

    rank = mpi_get_rank(comm);
    size = mpi_get_size(comm);
    if (rank < 0 || size < 0) {
        return MPI_ERR_COMM;
    }

    /* If not root, send data to the root. */
    if (rank != root) {
        ret = mpi_int_send(sbuf, count, dtype, root, REDUCE_TAG, comm, 0, &req);
        if (MPI_SUCCESS == ret) {
            ret = mpi_int_request_wait(&req, MPI_STATUS_IGNORE);
        }
        return ret;
    }

    /* Root receives and reduces messages.  Allocate buffer to receive
       messages. */
    if (size > 1) {
        buffer = (char*) malloc(dtype * count);
        if (NULL == buffer) return MPI_ERR_SYSRESOURCE;
    }

    /* Initialize the receive buffer. */
    if (rank == (size - 1)) {
        memcpy(rbuf, sbuf, count*dtype);
    } else {
        ret = mpi_int_recv(rbuf, count, dtype, size - 1,
                           REDUCE_TAG, comm, &req);
        if (MPI_SUCCESS != ret) goto cleanup;
        ret = mpi_int_request_wait(&req, MPI_STATUS_IGNORE);
        if (MPI_SUCCESS != ret) goto cleanup;
    }

    /* Loop receiving and calling reduction function (C or Fortran). */
    for (i = size - 2; i >= 0; --i) {
        if (rank == i) {
            inbuf = (char*)sbuf;
        } else {
            ret = mpi_int_recv(buffer, count, dtype, i, REDUCE_TAG, comm, 
                               &req);
            if (MPI_SUCCESS != ret) goto cleanup;
            inbuf = buffer;
        }

        /* Call reduction function. */
        switch(dtype) {
        case MPI_BYTE:
            ret = MPI_ERR_OP;
            break;
        case MPI_CHAR:
            ret = doOp(op, (char*)inbuf, (char*)rbuf, count);
            break;
        case MPI_SHORT:
            ret = doOp(op, (short*)inbuf, (short*)rbuf, count);
            break;
        case MPI_INT:
            ret = doOp(op, (int*)inbuf, (int*)rbuf, count);
            break;
        case MPI_LONG:
            ret = doOp(op, (long*)inbuf, (long*)rbuf, count);
            break;
        case MPI_FLOAT:
            ret = doOp(op, (float*)inbuf, (float*)rbuf, count);
            break;
        case MPI_DOUBLE:
            ret = doOp(op, (double*)inbuf, (double*)rbuf, count);
            break;
        default:
            ret = MPI_ERR_OP;
            break;
        }

        if (MPI_SUCCESS != ret) goto cleanup;
    }

    ret = MPI_SUCCESS;

 cleanup:
    if (buffer) free(buffer);
    return ret;
}


static inline int
mpi_int_bcast(void *buffer, int count, MPI_Datatype dtype, int root, MPI_Comm comm)
{
    if (MPI_COMM_SELF == comm) return MPI_SUCCESS;

    int rank = mpi_get_rank(comm);
    int ret;
    MPI_Request req;

    if (rank < 0) {
        return MPI_ERR_COMM;
    } else if (rank == root) {
        int size = mpi_get_size(comm);
    
        for (int r = 0; r < size; ++r) {
            if (r != rank) {
                ret = mpi_int_send(buffer, count, dtype, r, BCAST_TAG, comm, 0, &req);
                if (MPI_SUCCESS != ret) return ret;
                ret = mpi_int_request_wait(&req, MPI_STATUS_IGNORE);
                if (MPI_SUCCESS != ret)  return ret;
            }
        }
    } else {
        ret = mpi_int_recv(buffer, count, dtype, root, BCAST_TAG, 
                           comm, &req);
        if (MPI_SUCCESS != ret) return ret;
        ret = mpi_int_request_wait(&req, MPI_STATUS_IGNORE);
        if (MPI_SUCCESS != ret)  return ret;
    }

    return MPI_SUCCESS;
}


int
MPI_Allreduce(void *sbuf, void *rbuf, int count, MPI_Datatype dtype, 
              MPI_Op op, MPI_Comm comm)
{
    int ret;

    ret = mpi_int_reduce(sbuf, rbuf, count, dtype, op, 0, comm);
    if (MPI_SUCCESS == ret) {
        ret = mpi_int_bcast(rbuf, count, dtype, 0, comm);
    }

    return mpi_error_check("MPI_Allreduce", ret);
}


int
MPI_Barrier(MPI_Comm comm)
{
    int ret, rank, size;
    MPI_Request *reqs = NULL;
    MPI_Request req;

    if (MPI_COMM_SELF == comm) return MPI_SUCCESS;

    rank = mpi_get_rank(comm);
    size = mpi_get_size(comm);
    if (rank < 0 || size <= 0) return mpi_error_check("MPI_Barrier", MPI_ERR_COMM);

    if (0 == rank) {
        bool done;

        reqs = (MPI_Request*) malloc(sizeof(MPI_Request) * size);
        if (NULL == reqs) return mpi_error_check("MPI_Barrier", MPI_ERR_SYSRESOURCE);

        /* wait for everyone to ping us */
        done = false;
        for (int i = 1 ; i < size ; ++i) {
            ret = mpi_int_recv(NULL, 0, MPI_BYTE, i, BARRIER_TAG, comm, reqs + i);
            if (MPI_SUCCESS != ret) goto cleanup;
        }
        while (!done) {
            done = true;
            mpi_int_progress();
            for (int i = 1 ; i < size ; ++i) {
                if (MPI_REQUEST_NULL == reqs[i]) {
                    continue;
                } else if (false != reqs[i]->complete) {
                    ret = mpi_int_request_complete(reqs + i, MPI_STATUS_IGNORE, false);
                    if (MPI_SUCCESS != ret) goto cleanup;
                } else {
                    done = false;
                }
            }
        }

        /* ping everyone */
        done = false;
        for (int i = 1 ; i < size ; ++i) {
            ret = mpi_int_send(NULL, 0, MPI_BYTE, i, BARRIER_TAG, comm, 0, reqs + i);
            if (MPI_SUCCESS != ret) goto cleanup;
        }
        while (!done) {
            done = true;
            mpi_int_progress();
            for (int i = 1 ; i < size ; ++i) {
                if (MPI_REQUEST_NULL == reqs[i]) {
                    continue;
                } else if (0 != reqs[i]->complete) {
                    ret = mpi_int_request_complete(reqs + i, MPI_STATUS_IGNORE, false);
                    if (MPI_SUCCESS != ret) goto cleanup;
                } else {
                    done = false;
                }
            }
        }

    } else {
        ret = mpi_int_send(NULL, 0, MPI_BYTE, 0, BARRIER_TAG, comm, 0, &req);
        if (MPI_SUCCESS != ret) goto cleanup;
        ret = mpi_int_request_wait(&req, MPI_STATUS_IGNORE);
        if (MPI_SUCCESS != ret) goto cleanup;
        ret = mpi_int_recv(NULL, 0, MPI_BYTE, 0, BARRIER_TAG, comm, &req);
        if (MPI_SUCCESS != ret) goto cleanup;
        ret = mpi_int_request_wait(&req, MPI_STATUS_IGNORE);
        if (MPI_SUCCESS != ret) goto cleanup;
    }

    ret = MPI_SUCCESS;

 cleanup:
    if (NULL != reqs) free(reqs);
    return mpi_error_check("MPI_Barrier", ret);
}


int
MPI_Bcast(void *buffer, int count, MPI_Datatype dtype, int root, MPI_Comm comm)
{
    return mpi_error_check("MPI_Bcast", 
                           mpi_int_bcast(buffer, count, dtype, root, comm));
}


int
MPI_Reduce(void *sbuf, void *rbuf, int count, MPI_Datatype dtype, 
           MPI_Op op, int root, MPI_Comm comm)
{
    return mpi_error_check("MPI_Reduce",
                           mpi_int_reduce(sbuf, rbuf, count, dtype, op, root, comm));
}
