// Copyright 2009-2010 Sandia Corporation. Under the terms
// of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
// Government retains certain rights in this software.
// 
// Copyright (c) 2009-2010, Sandia Corporation
// All rights reserved.
// 
// This file is part of the SST software package. For license
// information, see the LICENSE file in the top level directory of the
// distribution.


#ifndef MPI_INTERNAL_H
#define MPI_INTERNAL_H

#include <stdio.h>

#include "mpi.h"
#include "netsim.h"
#include "pool.h"

#if defined(c_plusplus) || defined(__cplusplus)
extern "C" {
#endif

#if defined(DEBUG) && DEBUG != 0
#define DEBUG_PRINTF(a)        \
do {                           \
    printf a;                  \
    fflush(NULL);              \
} while (0)
#else
#define DEBUG_PRINTF(a)
#endif


struct _request {
    volatile bool complete;
    struct _status status;
};
typedef struct _request request;

/* table of type sizes */
extern int dtype_sizes[MPI_TYPE_MAX];
extern int mpi_comm_world_rank;
extern int mpi_comm_world_size;
extern int mpi_error_abort;

extern pool<request> *request_pool;

inline int
mpi_get_rank(MPI_Comm comm)
{
    if (MPI_COMM_WORLD == comm) {
        return mpi_comm_world_rank;
    } else if (MPI_COMM_SELF == comm) {
        return 0;
    } else {
        return -1;
    }
}


inline int
mpi_get_size(MPI_Comm comm)
{
    if (MPI_COMM_WORLD == comm) {
        return mpi_comm_world_size;
    } else if (MPI_COMM_SELF == comm) {
        return 1;
    } else {
        return -1;
    }
}


/* match/ignore bit manipulation
 *
 * 01234567 01234567 01234567 01234567 01234567 01234567 01234567 01234567
 *                  |                 |
 *      context id  |      source     |            message tag
 *                  |                 |
 */

#define MPI_SOURCE_MASK   0x0000FFFF00000000ULL
#define MPI_TAG_MASK      0x00000000FFFFFFFFULL

#define MPI_SOURCE_IGNR   MPI_SOURCE_MASK
#define MPI_TAG_IGNR      0x000000007FFFFFFFULL

/* get the tag from the bits */ 
#define MPI_GET_TAG(match_bits, tag)                \
{                                                   \
    tag = (int) (match_bits & MPI_TAG_MASK);        \
}


/* get the tag from the bits */ 
#define MPI_GET_SRC(match_bits, src)                    \
{                                                       \
    src = (int) ((match_bits & MPI_SOURCE_MASK) >> 32); \
}

/* send posting */
#define MPI_SET_SEND_BITS(match_bits, contextid, source, tag)           \
{                                                                       \
    match_bits = contextid;                                             \
    match_bits = (match_bits << 16);                                    \
    match_bits |= source;                                               \
    match_bits = (match_bits << 32);                                    \
    match_bits |= (MPI_TAG_MASK & tag);                                 \
}

/* receive posting */
#define MPI_SET_RECV_BITS(match_bits, ignore_bits, contextid, source, tag) \
{                                                                       \
    match_bits = 0;                                                     \
    ignore_bits = 0;                                                    \
                                                                        \
    match_bits = contextid;                                             \
    match_bits = (match_bits << 16);                                    \
                                                                        \
    if (MPI_ANY_SOURCE == source) {                                     \
        ignore_bits |= MPI_SOURCE_IGNR;                                 \
    } else {                                                            \
        match_bits |= source;                                           \
    }                                                                   \
    match_bits = (match_bits << 32);                                    \
                                                                        \
    if (MPI_ANY_TAG == tag) {                                           \
        ignore_bits |= MPI_TAG_IGNR;                                    \
    } else {                                                            \
        match_bits |= (MPI_TAG_MASK & tag);                             \
    }                                                                   \
}


inline void
mpi_int_progress(void)
{
    netsim_cqe_t cqe;
    int ret;

    ret = netsim_cq_poll(&cqe);
    if (0 == ret) {
        request *req = (request*)(long)cqe.user_data;
        req->status.MPI_ERROR = cqe.status;
        req->status.st_length = cqe.data_len;
        MPI_GET_SRC(cqe.send_match_bits, req->status.MPI_SOURCE);
        MPI_GET_TAG(cqe.send_match_bits, req->status.MPI_TAG);

        DEBUG_PRINTF(("%02d: mpi_int_progress:     %p, %d, %d, %d, %d\n",
                      mpi_comm_world_rank,
                      req, 
                      req->status.MPI_SOURCE,
                      req->status.MPI_TAG,
                      req->status.MPI_ERROR,
                      req->status.st_length));

        req->complete = true;
    }
}


inline int
mpi_int_request_complete(MPI_Request *req, MPI_Status *status, bool set_error)
{
    int ret = (*req)->status.MPI_ERROR;

    if (MPI_STATUS_IGNORE != status) {
        status->MPI_SOURCE = (*req)->status.MPI_SOURCE;
        status->MPI_TAG = (*req)->status.MPI_TAG;
        if (set_error) {
            status->MPI_ERROR = (*req)->status.MPI_ERROR;
        }
        status->st_length = (*req)->status.st_length;
    }
    DEBUG_PRINTF(("%02d: mpi_int_req_complete: %p, %d, %d, %d, %d\n",
                  mpi_comm_world_rank,
                  *req, 
                  (*req)->status.MPI_SOURCE,
                  (*req)->status.MPI_TAG,
                  (*req)->status.MPI_ERROR,
                  (*req)->status.st_length));

    request_pool->returnItem(*req);
    *req = MPI_REQUEST_NULL;

    return ret;
}


inline int
mpi_int_request_wait(MPI_Request *req, MPI_Status *status)
{
    while (1) {
        mpi_int_progress();
        if (0 != (*req)->complete) {
            return mpi_int_request_complete(req, status, false);
        }
    }
}


inline void
mpi_int_status_make_empty(MPI_Status *status) 
{
    status->MPI_SOURCE = MPI_ANY_SOURCE;
    status->MPI_TAG = MPI_ANY_TAG;
    status->MPI_ERROR = MPI_SUCCESS;
    status->st_length = 0;
}


inline int
mpi_error_check(const char *func, int return_code)
{
    if (MPI_SUCCESS != return_code && mpi_error_abort) {
        printf("MPI rank %d aborting in %s with code %d\n",
               mpi_comm_world_rank, func, return_code);
        abort();
    }

    return return_code;
}


inline int
mpi_int_send(void *buf, int count, int dtype,
             int dest, int tag, MPI_Comm comm, int sync,
             MPI_Request *reqp)
{
    uint64_t match_bits;
    int ret;
    int real_dest = (MPI_COMM_SELF == comm) ? mpi_comm_world_rank : dest;
    request * req = request_pool->getItem();
    if (NULL == req) return -1;
    
    req->complete = false;

    MPI_SET_SEND_BITS(match_bits, comm, mpi_comm_world_rank, tag);

    DEBUG_PRINTF(("%02d: mpi_int_send:         %d, %p, %d, %d, %#llx, %p\n", 
                  mpi_comm_world_rank,
                  real_dest, buf, count * dtype_sizes[dtype], sync,
                  match_bits, req));
    *reqp = req;
    ret = netsim_tx_start(real_dest, buf, count * dtype_sizes[dtype], sync, 
                           match_bits, (uint64_t)(long) req);
    if (0 != ret) {
        request_pool->returnItem(req);
        req = MPI_REQUEST_NULL;
    }

    return ret;
}


inline int
mpi_int_recv(void *buf, int count, int dtype,
             int source, int tag, MPI_Comm comm, 
             MPI_Request *reqp)
{
    uint64_t match_bits, ignore_bits;
    int ret;
    int real_source = (MPI_COMM_SELF == comm) ? mpi_comm_world_rank : source;
    request * req = request_pool->getItem();
    if (NULL == req) return -1;
    
    req->complete = false;

    MPI_SET_RECV_BITS(match_bits, ignore_bits, comm, real_source, tag);

    DEBUG_PRINTF(("%02d: mpi_int_recv:         %p, %d, %#llx, %#llx %p\n", 
                  mpi_comm_world_rank,
                  buf, count * dtype_sizes[dtype],
                  match_bits, ignore_bits, req));
    *reqp = req;
    ret = netsim_rx_start(buf, count * dtype_sizes[dtype],
                          match_bits, ignore_bits, (uint64_t)(long)req);
    if (0 != ret) {
        request_pool->returnItem(req);
        req = MPI_REQUEST_NULL;
    }

    return ret;
}


/* return is 0, 2, or error */
inline int
mpi_int_probe(int source, int tag, MPI_Comm comm, MPI_Status *status)
{
    uint64_t match_bits, ignore_bits;
    int ret;
    int real_source = (MPI_COMM_SELF == comm) ? mpi_comm_world_rank : source;
    request * req = request_pool->getItem();
    if (NULL == req) return -1;
    
    req->complete = false;

    MPI_SET_RECV_BITS(match_bits, ignore_bits, comm, real_source, tag);


    DEBUG_PRINTF(("%02d: mpi_int_probe: %#llx, %#llx, %p\n", 
                  mpi_comm_world_rank,
                  match_bits, ignore_bits, req));
    ret = netsim_probe_start(match_bits, ignore_bits, (uint64_t)(long)req);
    if (0 != ret) {
        request_pool->returnItem(req);
        req = MPI_REQUEST_NULL;
        return ret;
    }

    return mpi_int_request_wait(&req, status);
}


#if defined(c_plusplus) || defined(__cplusplus)
}
#endif


#endif
