// Copyright 2009-2010 Sandia Corporation. Under the terms
// of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
// Government retains certain rights in this software.
// 
// Copyright (c) 2009-2010, Sandia Corporation
// All rights reserved.
// 
// This file is part of the SST software package. For license
// information, see the LICENSE file in the top level directory of the
// distribution.


#include "mpi_internal.h"

#include <string.h>

int
MPI_Testall(int count, MPI_Request *array_reqs, int *flag, MPI_Status *array_status)
{
    int ret = MPI_SUCCESS;
    MPI_Status *status;

    mpi_int_progress();

    /* start by assuming there's an incomplete request.  If we find
       one, exit.  If not, jump to next lower loop */
    *flag = 0;
    for (int i = 0 ; i < count ; ++i) {
        if (MPI_REQUEST_NULL == array_reqs[i]) continue;
        if (0 == array_reqs[i]->complete) {
            return MPI_SUCCESS;
        }
    }

    /* all requests have completed, so fill in the status and free the
       requests */
    *flag = 1;
    ret = 0;
    for (int i  = 0 ; i < count ; ++i) {
        if (MPI_STATUSES_IGNORE == array_status) {
            status = MPI_STATUS_IGNORE;
        } else {
            status = array_status + i;
        }

        if (MPI_REQUEST_NULL == array_reqs[i]) {
            mpi_int_status_make_empty(status);
        } else {
            ret |= mpi_int_request_complete(array_reqs + i, status, true);
        }
    }

    if (MPI_SUCCESS != ret) ret = MPI_ERR_IN_STATUS;
    return mpi_error_check("MPI_Testall", ret);
}


int
MPI_Testany(int count, MPI_Request *array_reqs, int *index, int *flag, 
            MPI_Status *status)
{
    int ret = MPI_SUCCESS;

    /* check to see if any requests are complete.  Need to complete
       exactly one */
    *flag = 1; /* set to one for no active request case */
    *index = MPI_UNDEFINED;
    mpi_int_progress();
    for (int i = 0 ; i < count ; ++i) {
        if (MPI_REQUEST_NULL == array_reqs[i]) continue;
        *flag = 0; /* found an active request - assume incomplete */
        if (0 != array_reqs[i]->complete) {
            *flag = 1;
            *index = i;
            ret = mpi_int_request_complete(array_reqs + i, status, false);
            break;
        }
    }

    return mpi_error_check("MPI_Testany", ret);
}


int
MPI_Test(MPI_Request *req, int *done, MPI_Status *status)
{
    int ret = MPI_SUCCESS;

    if (MPI_REQUEST_NULL == *req) {
        mpi_int_status_make_empty(status);
        *done = 1;
        return MPI_SUCCESS;
    }

    mpi_int_progress();

    if (0 != (*req)->complete) {
        *done = 1;
        ret = mpi_int_request_complete(req, status, false);
    } else {
        *done = 0;
    }

    return mpi_error_check("MPI_Test", ret);
}


int
MPI_Testsome(int count, MPI_Request *array_reqs, int *outcount, int *array_idx, 
             MPI_Status *array_status)
{
    *outcount = 0;
    int ret = MPI_SUCCESS;
    bool found_active = false;
    MPI_Status *status;

    mpi_int_progress();
    for (int i = 0 ; i < count ; ++i) {
        if (MPI_REQUEST_NULL == array_reqs[i]) continue;
        found_active = true;
        if (0 != array_reqs[i]->complete) {
            if (MPI_STATUSES_IGNORE == array_status) {
                status = MPI_STATUS_IGNORE;
            } else {
                status = array_status + *outcount;
            }
            ret |= mpi_int_request_complete(array_reqs + i, 
                                            status,
                                            true);
            array_idx[*outcount] = i;
            (*outcount)++;
        }
    }

    if (!found_active) {
        *outcount = MPI_UNDEFINED;
        return MPI_SUCCESS;
    }

    if (MPI_SUCCESS != ret) ret = MPI_ERR_IN_STATUS;
    return mpi_error_check("MPI_Testsome", ret);
}


int
MPI_Waitall(int count, MPI_Request *array_reqs, MPI_Status *array_status)
{
    bool done = true;
    int ret = MPI_SUCCESS;
    MPI_Status *status;

    for (int i = 0 ; i < count ; ++i) {
        if (MPI_STATUSES_IGNORE == array_status) {
            status = MPI_STATUS_IGNORE;
        } else {
            status = array_status + i;
        }

        if (MPI_REQUEST_NULL == array_reqs[i]) {
            mpi_int_status_make_empty(status);
        } else if (0 != array_reqs[i]->complete) {
            ret |= mpi_int_request_complete(array_reqs + i, status, true);
        } else {
            done = false;
        }
    }

    while (!done) {
        mpi_int_progress();
        for (int i = 0 ; i < count ; ++i) {
            if (MPI_STATUSES_IGNORE == array_status) {
                status = MPI_STATUS_IGNORE;
            } else {
                status = array_status + i;
            }

            if (MPI_REQUEST_NULL == array_reqs[i]) {
                continue;
            } else if (0 != array_reqs[i]->complete) {
                ret |= mpi_int_request_complete(array_reqs + i, status, true);
            } else {
                done = false;
            }
        }
    }

    if (MPI_SUCCESS != ret) ret = MPI_ERR_IN_STATUS;
    return mpi_error_check("MPI_Waitall", ret);
}


int
MPI_Waitany(int count, MPI_Request *array_reqs, int *index, MPI_Status *status)
{
    while (1) {
        bool found_active = false;

        mpi_int_progress();
        for (int i = 0 ; i < count ; ++i) {
            if (MPI_REQUEST_NULL == array_reqs[i]) continue;
            found_active = true;
            if (0 != array_reqs[i]->complete) {
                *index = i;
                return mpi_error_check("MPI_Waitany", 
                          mpi_int_request_complete(array_reqs + i, status, false));
            }
        }

        if (!found_active) {
            *index = MPI_UNDEFINED;
            return MPI_SUCCESS;
        }
    }
}


int
MPI_Wait(MPI_Request *req, MPI_Status *status)
{
    if (MPI_REQUEST_NULL == *req) {
        mpi_int_status_make_empty(status);
        return MPI_SUCCESS;
    }

    while (0 == (*req)->complete) {
        mpi_int_progress();
    }

    return mpi_error_check("MPI_Wait", 
                           mpi_int_request_complete(req, status, false));
}


int
MPI_Waitsome(int count, MPI_Request *array_reqs, int *outcount, int *array_idx, 
             MPI_Status *array_status)
{
    *outcount = 0;
    int ret = 0;
    MPI_Status *status;

    while (0 == *outcount ) {
        bool found_active = false;

        mpi_int_progress();
        for (int i = 0 ; i < count ; ++i) {
            if (MPI_REQUEST_NULL == array_reqs[i]) continue;
            found_active = true;
            if (0 != array_reqs[i]->complete) {
                if (MPI_STATUSES_IGNORE == array_status) {
                    status = MPI_STATUS_IGNORE;
                } else {
                    status = array_status + *outcount;
                }
                ret |= mpi_int_request_complete(array_reqs + i, 
                                                status,
                                                true);
                array_idx[*outcount] = i;
                (*outcount)++;
            }
        }

        if (!found_active) {
            *outcount = MPI_UNDEFINED;
            return MPI_SUCCESS;
        }
    }

    if (MPI_SUCCESS != ret) ret = MPI_ERR_IN_STATUS;
    return mpi_error_check("MPI_Waitsome", ret);
}
