// Copyright 2009-2010 Sandia Corporation. Under the terms
// of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
// Government retains certain rights in this software.
// 
// Copyright (c) 2009-2010, Sandia Corporation
// All rights reserved.
// 
// This file is part of the SST software package. For license
// information, see the LICENSE file in the top level directory of the
// distribution.


/*
** A netsim API library with syscalls into a PowerPC
*/
#include <stdio.h>
#include <string.h>		// For memset()
#include <malloc/malloc.h>	// For malloc_printf()
#define malloc_printf	printf
#include "netsim.h"
#include "netsim_internal.h"
#include "ppc_syscall.h"

// These convert a #define number into a string constant
// We use this in the asm calls
#define STRINGIFY(x) #x
#define TOSTRING(x) STRINGIFY(x)
#define NETSIM_SYS_ENTER_STR TOSTRING(NETSIM_SYS_ENTER)
#define NETSIM_TX_ENTER_STR TOSTRING(NETSIM_TX_ENTER)
#define NETSIM_PICKUP_STR TOSTRING(NETSIM_SYS_PICKUP)


// Cached values
int cached_my_rank= -1;
int cached_nranks= -1;
int netsim_initialized= 0;


//
// Some common macros and functions
//
__inline__ void
trap(netsim_routine_t routine, netsim_params_t *params, int params_length)
{

int result;


    asm volatile ( 
	"mr r3, %1\n\t"				// Which netsim function to call
	"mr r4, %2\n\t"				// Address of shared data struct
	"mr r5, %3\n\t"				// Size of shared data struct
	"li r0, "NETSIM_SYS_ENTER_STR"\n\t"	// Set to syscall
	"sc\n\t"				// make the sys call
	"mr %0, r3"				// collect results
	: "=r" (result)				// Output operands
	: "r" (routine),			// Input operands
	  "r" (params),
	  "r" (params_length)
	: "r3", "r4", "r5"			// clobbered registers
        );
}  // end of trap()



// Send trap
__inline__ void
trap_tx(netsim_routine_t routine, netsim_params_t *params, int params_length,
    void *buf, int buf_len)
{

int result;


    asm volatile ( 
	"mr r3, %1\n\t"				// Which netsim function to call
	"mr r4, %2\n\t"				// Address of paramater struct
	"mr r5, %3\n\t"				// Size of paramater struct
	"mr r6, %4\n\t"				// Address of buffer
	"mr r7, %5\n\t"				// Length of buffer
	"li r0, "NETSIM_TX_ENTER_STR"\n\t"	// Set to syscall
	"sc\n\t"				// make the sys call
	"mr %0, r3"				// collect results
	: "=r" (result)				// Output operands
	: "r" (routine),			// Input operands
	  "r" (params),
	  "r" (params_length),
	  "r" (buf),
	  "r" (buf_len)
	: "r3", "r4", "r5", "r6", "r7"		// clobbered registers
        );
}  // end of trap()



__inline__ void
pickup(netsim_params_t *params, int params_length)
{

int result;


    asm volatile ( 
	"mr r4, %1\n\t"				// Address of shared data struct
	"mr r5, %2\n\t"				// Size of shared data struct
	"li r0, "NETSIM_PICKUP_STR"\n\t"	// Set to syscall
	"sc\n\t"				// make the sys call
	"mr %0, r3"				// collect results
	: "=r" (result)				// Output operands
	: "r" (params),				// Input operands
	  "r" (params_length)
	: "r3", "r4", "r5"			// clobbered registers
        );
}  // end of pickup()



#define done() \
    if (result == 1)   { \
        return NETSIM_SUCCESS; \
    } else   { \
        return NETSIM_ERROR; \
    }



/*
** Finalize netsim interface.
** May be called exactly once.
*/
int
netsim_finalize(void)
{

static int first_time= 1;
int result;
netsim_params_t params;


    if (!first_time)   {
        malloc_printf("netsim_finalize() should not be called more than once!\n");
        return NETSIM_ERROR;
    }
    first_time= 0;

    trap(NETSIM_FINALIZE, &params, sizeof(params));
    result= params.rc;
    done();

}  /* end of netsim_finalize() */



/*
** Because netsim_size() and netsim_rank() may be called before netsim_init()
** we call this function from the first function that needs the init information.
*/
static int
_do_netsim_init(void)
{

int result;
netsim_params_t params;


    // init doesn't really use match bits. We set them for testing purposes.
    memset(&params, 0, sizeof(params));
    params.match_bits= 0x123456789abcdef0ll;

    trap(NETSIM_INIT, &params, sizeof(params));
    result= params.rc;

    // Wait for the NIC to respond
    params.rc= 0;
    while (!params.rc)   {
	pickup(&params, sizeof(params));
    }

    // Store some cached values
    cached_my_rank= params.my_rank;
    cached_nranks= params.nranks;
    netsim_initialized= 1;
    done();

}  /* end of _do_netsim_init() */



/*
** Initialize netsim interface.
** May be called multiple times.
*/
int
netsim_init(void)
{

static int first_time= 1;


    if (!first_time)   {
	/* Been here already */
	return NETSIM_SUCCESS;
    } else   {
	first_time= 0;
    }

    if (netsim_initialized)   {
	// Already done!
	return NETSIM_SUCCESS;
    }

    return _do_netsim_init();

}  /* end of netsim_init() */



/*
** Number of processes in the parallel job.
** May be called before netsim_init.
*/
int
netsim_size(void)
{

    if (!netsim_initialized)   {
	_do_netsim_init();
    }

    // malloc_printf("netsim_size() returning %d\n", cached_nranks);
    return cached_nranks;

}  // end of netsim_size()
  


/*
** My rank in the parallel job.
** May be called before netsim_init.
*/
int
netsim_rank(void)
{

    if (!netsim_initialized)   {
	_do_netsim_init();
    }

    // malloc_printf("netsim_rank() returning %d\n", cached_my_rank);
    return cached_my_rank;

}  // end of netsim_rank()



/*
** Start a send
*/
int
netsim_tx_start(uint32_t dest, void *buf, int msgSize,
                    int synchronous, uint64_t match_bits,
                    uint64_t user_data)
{

int result;
netsim_params_t params;


    if (synchronous)   {
	malloc_printf("Synchronous sends are not supported yet!\n");
	return NETSIM_ERROR;
    }

    params.dest= dest;
    params.buf= (uint64_t)buf;
    params.msgSize= msgSize;
    params.synchronous= 0;
    params.match_bits= match_bits;
    params.user_data= user_data;
    params.ignore_bits= 0;

    trap_tx(NETSIM_TX_START, &params, sizeof(params), buf, msgSize);
    result= params.rc;
    done();

}  // end of netsim_tx_start()



/*
** Post a recv
*/
int
netsim_rx_start(void *buf, int msgSize, 
                    uint64_t match_bits, uint64_t ignore_bits, 
                    uint64_t user_data)
{

int result;
netsim_params_t params;


    params.dest= -1;
    params.buf= (uint64_t)buf;
    params.msgSize= msgSize;
    params.match_bits= match_bits;
    params.ignore_bits= ignore_bits;
    params.user_data= user_data;

    trap(NETSIM_RX_START, &params, sizeof(params));
    result= params.rc;
    done();

}  // end of netsim_rx_start()



/*
** Start a probe
*/
int
netsim_probe_start(uint64_t match_bits, uint64_t ignore_bits, 
                       uint64_t user_data)
{

int result;
netsim_params_t params;


    params.match_bits= match_bits;
    params.ignore_bits= ignore_bits;
    params.user_data= user_data;

    trap(NETSIM_PROBE_START, &params, sizeof(params));
    result= params.rc;
    done();

}  // end of netsim_probe_start()



/*
** Return a completion queue entry.
** Returns -1 if empty
*/
int
netsim_cq_poll(netsim_cqe_t *cqe)
{

int result;
netsim_params_t params;


    pickup(&params, sizeof(params));
    result= params.rc;

    if (result)   {
	if (params.rc)   {
	    cqe->status= 0;
	} else   {
	    cqe->status= 1;
	}
	cqe->data_len= params.msgSize;
	cqe->send_match_bits= params.match_bits;
	cqe->user_data= params.user_data;
	return NETSIM_SUCCESS;
    } else   {
	return -1;  // empty
    }

}  // end of netsim_cq_poll()



/*
** get time (clockcycles)
*/ 
#include <sys/time.h>
double
netsim_get_clock(void)
{

int rc;
struct timeval tp;
void *tzp;


    rc= gettimeofday(&tp, &tzp);
    // malloc_printf("secs %d, usecs %d\n", (int)tp.tv_sec, tp.tv_usec);
    return (double)tp.tv_sec + (double)(tp.tv_usec / 1000000.0);

}  // end of netsim_get_clock()
