// Copyright 2009-2010 Sandia Corporation. Under the terms
// of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
// Government retains certain rights in this software.
// 
// Copyright (c) 2009-2010, Sandia Corporation
// All rights reserved.
// 
// This file is part of the SST software package. For license
// information, see the LICENSE file in the top level directory of the
// distribution.


/*
** $Id:$
** Basically a many-to-one test to exercise CAM overflow routines.
** Node 0 sets up (p - 1) * m receives and (p - 1) nodes send m
** messages each to node 0. This is repeated n times. Default for
** m = 1, and n = 10.
** If node 0 gets fewer messages than it expects during each iteration,
** it sends itself messages to clear out the pre-posted receives.
**
** Added the following command line options:
**     -l1 <len>	Set L1 message length to <len>
**     -l2 <len>	Set L2 message length to <len>
**     -Y <n>		How many nodes will send messages of length L1 
**			The remaning nodes will send messages of length L2
**
** April 2005, Rolf Riesen, Sandia National Laboratories
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include "mpi.h"


/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */
/*
** Functions
*/
static void init_bufs(int **sbuf, int ***rbuf, int L1, int L2, int m);
static void run_tests(int *sbuf, int **rbuf, int L1, int L2, int Y, int m, int n, int waittime);
static void prepostrcv(int **rbuf, int size, int m, MPI_Request *rreq);
static void sends(int *sbuf, int L1, int L2, int Y, int m);
static MPI_Request *alloc_rreq(int m);
static void collect_msg(int m, int waittime, MPI_Request *rreq);
static void usage(char *pname);



/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */
/*
** Globals
*/
static int debug;
static int rank, nnodes;



/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */
/*
** Constants
*/
#define NODE0		(0)
#define DEFAULT_SIZE	(4 * 1024)
#define DEFAULT_M	(1)
#define DEFAULT_N	(10)
#define DEFAULT_T	(1)
#define TAG		(511)
#define TRUE		(1)



/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */
/*
** Command Line Options
*/
static struct option long_options[] = {
    {"v", no_argument, NULL, 'v'},
    {"t", required_argument, NULL, 't'},
    {"m", required_argument, NULL, 'm'},
    {"n", required_argument, NULL, 'n'},
    {"l1", required_argument, NULL, '1'},
    {"l2", required_argument, NULL, '2'},
    {"Y", required_argument, NULL, 'y'},
    {0, 0, 0, 0}
};



/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

int
main(int argc, char **argv)
{

extern char *optarg;
extern int opterr;
int ch;
int option_index = 0;

int L1, L2, Y;
int *sbuf, **rbuf;
int m, n;
int waittime;


    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &nnodes);


    /* Set defaults */
    L1= DEFAULT_SIZE;
    L2= DEFAULT_SIZE;
    Y= nnodes - 1;
    m= DEFAULT_M;
    n= DEFAULT_N;
    waittime= DEFAULT_T;
    debug= 0;
    if (rank != NODE0)   {
	opterr= 0;
    }


    /* Process the command line */
    while ((ch= getopt_long(argc, argv, "", long_options, &option_index)) != EOF)   {
        switch (ch)   {
            case 'v': debug++;
		      break;
            case 't': waittime= strtol(optarg, NULL, 10);
                      break;
            case 'm': m= strtol(optarg, NULL, 10);
                      break;
            case 'n': n= strtol(optarg, NULL, 10);
                      break;
            case '1': L1= strtol(optarg, NULL, 10);
                      break;
            case '2': L2= strtol(optarg, NULL, 10);
                      break;
            case 'y': Y= strtol(optarg, NULL, 10);
                      if ((Y < 0) || (Y > nnodes - 1))   {
			if (rank == NODE0)   {
			    fprintf(stderr, "Y must be >= 0 and <= p - 1!\n");
			    usage(argv[0]);
			}
			exit(-1);
		      }
                      break;
	    case '?':
            default:  if (rank == NODE0) usage(argv[0]);
                      exit(-1);
        }
    }


    L1= (L1 + 1) / sizeof(int); /* convert to ints */
    L2= (L2 + 1) / sizeof(int); /* convert to ints */
    if (rank == NODE0)   {
	printf("Testing CAM overflows\n");
	printf("===========================================================\n");
	printf("Message length L1:                          %ld bytes\n", L1 * (long int)sizeof(int));
	printf("Message length L2:                          %ld bytes\n", L2 * (long int)sizeof(int));
	printf("Wait time is:                               %d second(s)\n", waittime);
	printf("Iterations:                                 %d\n", n);
	printf("Num messages (m):                           %d\n", m);
	printf("Num senders using size L1:                  %d\n", Y);
	printf("Num senders using size L2:                  %d\n", nnodes - 1 - Y);
	printf("Total messages per iteration (p - 1) * m:   %d\n", (nnodes - 1) * m);
	fflush(stdout);
    }


    if (debug > 1)   {
	printf("Node %2d: initializing buffers\n", rank);
	fflush(stdout);
    }
    init_bufs(&sbuf, &rbuf, L1, L2, m);

    if (debug > 1)   {
	printf("Node %2d: Starting tests\n", rank);
	fflush(stdout);
    }
    run_tests(sbuf, rbuf, L1, L2, Y, m, n, waittime);

    if (debug > 1)   {
	printf("Node %2d: done\n", rank);
	fflush(stdout);
    }

    MPI_Finalize();

    exit(0);

}  /* end of main() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

/*
** size is in ints
*/
static void
init_bufs(int **sbuf, int ***rbuf, int L1, int L2, int m)
{

int i;
int size;
int *snd_buf= NULL;
int **rcv_buf= NULL;


    /*
    ** We're being lazy and setup all receiver buffers to the larger
    ** message size: max(L1, L2)
    */
    if (L1 > L2)   {
	size= L1;
    } else   {
	size= L2;
    }



    if (rank == NODE0)   {
	/* Node 0 needs (p - 1) * m receive buffers */
	rcv_buf= (int **)malloc(((nnodes - 1) * m) * sizeof(int *));
	if (rcv_buf == NULL)   {
	    fprintf(stderr, "FAILED: Node %02d: Out of memory! [1] size %d\n",
		rank, ((nnodes - 1) * m) * (int)sizeof(int *));
	    exit(-1);
	}

	for (i= 0; i < (nnodes - 1) * m; i++)   {
	    rcv_buf[i]= (int *)malloc(size * sizeof(int));
	    if (rcv_buf[i] == NULL)   {
		fprintf(stderr, "FAILED: Node %02d: Out of memory! [2] size %d\n",
		    rank, size * (int)sizeof(int));
		exit(-1);
	    }
	    memset(rcv_buf[i], 0, size * sizeof(int));
	}
	if (debug)   {
	    printf("Node %2d: Initalized %d rcv bufs of len %ld\n", rank, (nnodes - 1) * m,
		(long int)(size * (int)sizeof(int)));
	    fflush(stdout);
	}
    }


    /* Everybody sets up a send buffer */
    snd_buf= (int *)malloc(size * sizeof(int));
    if (snd_buf == NULL)   {
	fprintf(stderr, "FAILED: Node %02d: Out of memory! [3] size %d\n", rank, size * (int)sizeof(int));
	exit(-1);
    }

    /* Fill it with data */
    for (i= 0; i < size; i++)   {
	snd_buf[i]= i;
    }

    *sbuf= snd_buf;
    *rbuf= rcv_buf;

    MPI_Barrier(MPI_COMM_WORLD);

}  /* end of init_bufs() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

static MPI_Request *
alloc_rreq(int m)
{

MPI_Request *rreq= NULL;


    rreq= (MPI_Request *)malloc((nnodes - 1) * m * sizeof(MPI_Request));
    if (rreq == NULL)   {
	fprintf(stderr, "FAILED: Node %02d: Out of memory! [4] size %d\n",
	    rank, (nnodes - 1) * m * (int)sizeof(MPI_Request));
	exit(-1);
    }

    if (debug)   {
	printf("Node %2d: rreq %p, num %d * size %ld = %ld, end %p\n", rank, rreq,
	    ((nnodes - 1) * m), (long int)sizeof(MPI_Request),
	    (long int)((nnodes - 1) * m * (int)sizeof(MPI_Request)),
	    rreq + ((nnodes - 1) * m));
	fflush(stdout);
    }


    memset(rreq, 0, (nnodes - 1) * m * sizeof(MPI_Request));
    return rreq;

}  /* end of alloc_rreq() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

static void
collect_msg(int m, int t, MPI_Request *rreq)
{

MPI_Status ignore;
int i, j;
int cnt;
int remain;
int good;
int flag;




    /* Give non-progressing libraries a chance to catch up */
    if (debug)   {
	printf("Node %2d: Do first round of checks for %d messages\n", rank, (nnodes - 1) * m);
	fflush(stdout);
    }

    for (i= 0; i < t; i++)   {
	sleep(1);
	for (j= 0; j < (nnodes - 1) * m; j++)   {
	    MPI_Test(&(rreq[j]), &flag, &ignore);
	}
    }


    /* Check how many we got */
    if (debug)   {
	printf("Node %2d: Count how many we actually got...\n", rank);
	fflush(stdout);
    }

    good= 0;
    for (j= 0; j < (nnodes - 1) * m; j++)   {
	MPI_Test(&(rreq[j]), &flag, &ignore);
	if (flag)   {
	    good++;
	}
    }

    remain= (nnodes - 1) * m - good;
    if (debug)   {
	printf("Node %2d: Found %d messages. Will clear out %d\n", rank, good, remain);
	fflush(stdout);
    }

    /* Now cancel the rest */
    cnt= 0;
    for (j= 0; j < (nnodes - 1) * m; j++)   {
	MPI_Test(&(rreq[j]), &flag, &ignore);
	if (!flag)   {
	    cnt++;
	    /* MPI lite doesn't support MPI_Cancel */
	    /* MPI_Cancel(&(rreq[j])); */
	    rreq[j]= MPI_REQUEST_NULL;
	}
    }

    if (debug)   {
	printf("Node %2d: Had to cancel %d recvs\n", rank, cnt);
	fflush(stdout);
    }

    /* Do the actual receives */
    for (j= 0; j < (nnodes - 1) * m; j++)   {
	if (rreq[j] != MPI_REQUEST_NULL)   {
	    /* Only the ones that were not cancelled */
	    MPI_Wait(&(rreq[j]), &ignore);
	}
    }

    if (debug)   {
	printf("Node %2d: Did %d receives\n", rank, j);
	fflush(stdout);
    }

    if (good == (nnodes - 1) * m)   {
	printf("passed:    %3d: Got %3d messages. Expected %3d\n", j, good, (nnodes - 1) * m);
	fflush(stdout);
    } else   {
	printf("FAILED:    %3d: Got %3d messages. Expected %3d\n", j, good, (nnodes - 1) * m);
	fflush(stdout);
    }

}  /* end of collect_msg() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

static void
run_tests(int *sbuf, int **rbuf, int L1, int L2, int Y, int m, int n, int waittime)
{

MPI_Request *rreq= NULL;
int size;
int i;


    if (rank == NODE0)   {
	rreq= alloc_rreq(m);
	printf("\n           Iter      Received           Wanted\n");
    }


    for (i= 0; i < n; i++)   {
	if (rank == NODE0)   {
	    /*
	    ** We're being lazy and setup all receiver buffers to the larger
	    ** message size: max(L1, L2)
	    */
	    if (L1 > L2)   {
		size= L1;
	    } else   {
		size= L2;
	    }

	    prepostrcv(rbuf, size, m, rreq);
	    if (debug)   {
		printf("Node %2d: preposted receives!\n", rank);
		fflush(stdout);
	    }

	    MPI_Barrier(MPI_COMM_WORLD);
	    if (debug)   {
		printf("Node %2d: Out of barrier. Going to sleep for %d seconds\n", rank, waittime);
		fflush(stdout);
	    }

	    /* Go receive the messages */
	    collect_msg(m, waittime, rreq);

	} else   {
	    if (debug > 1)   {
		printf("Node %2d: Waiting for node 0!\n", rank);
		fflush(stdout);
	    }
	    MPI_Barrier(MPI_COMM_WORLD);

	    /* Do all the sends */
	    sends(sbuf, L1, L2, Y, m);
	}
    }

    MPI_Barrier(MPI_COMM_WORLD);

}  /* end of run_tests() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

static void
prepostrcv(int **rbuf, int size, int m, MPI_Request *rreq)
{

int i;


    for (i= 0; i < (nnodes - 1) * m; i++)   {
	memset(rbuf[i], 0, size * sizeof(int));
	MPI_Irecv(rbuf[i], size, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &(rreq[i]));
	if (debug > 2)   {
	    printf("Node %2d: Posted buffer %3d. rreq %p\n", rank, i, rreq[i]);
	    fflush(stdout);
	}
    }

}  /* end of prepostrcv() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

static void
sends(int *sbuf, int L1, int L2, int Y, int m)
{

int i;
int size;
static MPI_Request *sreq= NULL;
static MPI_Status *ignore= NULL;


    if (sreq == NULL)   {
	sreq= (MPI_Request *)malloc(m * sizeof(MPI_Request));
	if (sreq == NULL)   {
	    fprintf(stderr, "FAILED: Node %02d: Out of memory! [5] size %d\n",
		rank, m * (int)sizeof(MPI_Request));
	    exit(-1);
	}
    }

    if (ignore == NULL)   {
	ignore= (MPI_Status *)malloc(m * sizeof(MPI_Status));
	if (ignore == NULL)   {
	    fprintf(stderr, "FAILED: Node %02d: Out of memory! [5] size %d\n", rank,
		m * (int)sizeof(MPI_Status));
	    exit(-1);
	}
    }

    if (rank <= Y)   {
	size= L1;
    } else   {
	size= L2;
    }

    for (i= 0; i < m; i++)   {
	MPI_Isend(sbuf, size, MPI_INT, NODE0, TAG, MPI_COMM_WORLD, &(sreq[i]));
	if (debug > 2)   {
	    printf("Node %2d: sending %ld bytes to node 0\n", rank,
		size * (long int)sizeof(int));
	    fflush(stdout);
	}
    }

    if (debug > 1)   {
	printf("Node %2d: Waiting for %d sends to complete...\n", rank, i);
	fflush(stdout);
    }

    MPI_Waitall(m, sreq, ignore);

}  /* end of sends() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

static void
usage(char *pname)
{

    fprintf(stderr, "Usage: %s [-n <n>] [-m <m>] [-t <t>] ", pname);
    fprintf(stderr, "[-l1 <l1>] [-l2 <l2>] [-Y <y>] [-v]\n");
    fprintf(stderr, "    <m>    How many messages each node sends (default %d)\n", DEFAULT_M);
    fprintf(stderr, "    <n>    Number of times to repeat the test (default %d)\n", DEFAULT_N);
    fprintf(stderr, "    <t>    How long to wait for messages (default %d)\n", DEFAULT_T);
    fprintf(stderr, "    <l2>   Set L1 message length to <l1> (default %d)\n", DEFAULT_SIZE);
    fprintf(stderr, "    <l2>   Set L2 message length to <l2> (default %d)\n", DEFAULT_SIZE);
    fprintf(stderr, "    <y>    How many nodes will send messages of length L1  (default p - 1 nodes (all but 0))\n");
    fprintf(stderr, "           The remaning nodes will send messages of length L2\n");
    fprintf(stderr, "    -v     Increase verbosity (can be repeated for more output)\n");
    fprintf(stderr, "\nFAILED\n");

}  /* end of usage() */
