// Copyright 2009-2010 Sandia Corporation. Under the terms
// of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
// Government retains certain rights in this software.
// 
// Copyright (c) 2009-2010, Sandia Corporation
// All rights reserved.
// 
// This file is part of the SST software package. For license
// information, see the LICENSE file in the top level directory of the
// distribution.


/*
** $Id:$
** Check unaligned buffer transfers
** This is a 2-node program.
**
** March 2005, Rolf Riesen, Sandia National Laboratories
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "mpi.h"


/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */
/*
** Functions
*/
void *mpi_malloc(size_t sz, int nnodes, int rank);
void init_bufs(unsigned int *seed, char **snd_aligned, char **rcv_aligned,
    char **cmp_aligned);
void run_tests(unsigned int seed, char *snd_aligned, char *rcv_aligned,
    char *cmp_aligned);
void fill_buf(long int seed, char *buf, int size);
void check_buf(long int seed, char *buf, int size);
void doit(unsigned int seed, char *cmp, char *snd, char *rcv, int len);
void run1(unsigned int seed, char *snd, char *rcv, char *cmp, int len);



/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */
/*
** Constants
*/
#define EightMB		(1 << 23)
#define BLOCKS		(5)		/* How many 8MB blocks we will transfer */
#define	INT64		((long int))	/* Make this the same size as a pointer */
#define NODE0		(0)
#define NODE1		(1)
#define TAG		(0x555)
#define R_SEED		(193746)
#define S_SEED		(213632213)


/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */
#define DEBUG
#undef DEBUG
int debug= 0;


int
main(int argc, char **argv)
{

char *snd_aligned;
char *rcv_aligned;
char *cmp_aligned;

unsigned int seed;
int rank;


    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    if (rank == NODE0)   {
	printf("Checking Unaligned Transfer Correctnes\n");
	printf("======================================\n");
    }

    init_bufs(&seed, &snd_aligned, &rcv_aligned, &cmp_aligned);

    run_tests(seed, snd_aligned, rcv_aligned, cmp_aligned);

    MPI_Barrier(MPI_COMM_WORLD);

    if (rank == NODE0)   {
	printf("PASSED\n");
    }
    MPI_Finalize();

    exit(0);

}  /* end of main() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

void *
mpi_malloc(size_t sz, int nnodes, int rank)
{

void *rv;
int ok;
int allok;


    /* Initialize these two values */
    ok= 1;
    rv= NULL;

    /* Attempt to allocate memory if a positive size was given */
    rv= malloc(sz);
    if (rv == NULL)   {
	/* Record that the malloc failed */
	ok= -1;
    }

    /* Check how all the malloc's did */
    MPI_Allreduce (&ok, &allok, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);

    if (allok < 0)   {
        /* Somebody's malloc failed... exit */
        if (rv != NULL)   {
            free (rv);
        }
        if (rank == NODE0)   {
            fprintf(stderr, "FAILED: Out of memory\n");
        }
        exit(-1);
    }

    return rv;

}  /* en dof mpi_malloc() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

void
init_bufs(unsigned int *seed, char **snd_aligned, char **rcv_aligned,
    char **cmp_aligned)
{

char *snd_buf;
char *rcv_buf;
char *cmp_buf;
int rank, nnodes;


    MPI_Comm_size(MPI_COMM_WORLD, &nnodes);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    /*
    ** We get a large buffer plus an extra 8MB so we can align it on a 8 MB boundary!
    */
    snd_buf= mpi_malloc((BLOCKS + 1) * EightMB, nnodes, rank);
    rcv_buf= mpi_malloc((BLOCKS + 1) * EightMB, nnodes, rank);
    cmp_buf= mpi_malloc((BLOCKS + 1) * EightMB, nnodes, rank);


    /* Align the buffers */
    if (debug)   {
	printf("    %2d: snd_buf %p\n", rank, snd_buf);
    }
    *snd_aligned= (char *) ((long int) (snd_buf + EightMB) & ~((long int)EightMB - 1));
    if (debug)   {
	printf("    %2d: snd_aligned  %p\n", rank, snd_aligned);
	printf("    %2d: rcv_buf %p\n", rank, rcv_buf);
    }
    *rcv_aligned= (char *)((long int)(rcv_buf + EightMB) & ~(EightMB - 1));
    if (debug)   {
	printf("    %2d: rcv_aligned  %p\n", rank, rcv_aligned);
    }
    *cmp_aligned= (char *)((long int)(cmp_buf + EightMB) & ~(EightMB - 1));
    if (debug)   {
	printf("    %2d: cmp_aligned  %p\n", rank, cmp_aligned);
    }


    /* Fill the send and receive buffer with random date */
    if (rank == NODE0)   {
	printf("    Filling buffers with random data\n");
    }

    if (rank == NODE0)   {
	/* Genreate and broadcast random seed */
	*seed= (int)time(NULL) * 117;
	MPI_Bcast(seed, 1, MPI_LONG, 0, MPI_COMM_WORLD);
    } else   {
	MPI_Bcast(seed, 1, MPI_LONG, 0, MPI_COMM_WORLD);
    }
    if (debug)   {
	printf("    %2d: seed is  %d\n", rank, *seed);
    }

    fill_buf(*seed + S_SEED, *snd_aligned, BLOCKS * EightMB);
    fill_buf(*seed + R_SEED, *rcv_aligned, BLOCKS * EightMB);

    if (rank == NODE0)   {
	printf("    Checking buffers... \n");
    }

    check_buf(*seed + S_SEED, *snd_aligned, BLOCKS * EightMB);
    check_buf(*seed + R_SEED, *rcv_aligned, BLOCKS * EightMB);

}  /* end of init_bufs() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

void
fill_buf(long int seed, char *buf, int size)
{

int i;


    srandom(seed);
    for (i = 0; i < size; i++) {
	buf[i]= (int)random();
    }

}  /* end of fill_buf() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

void
check_buf(long int seed, char *buf, int size)
{

int i;
char ch;
int rank;


    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    srandom(seed);
    for (i = 0; i < size; i++) {
	ch = (int)random();
	if (buf[i] != ch)   {
	    fprintf(stderr, "    %2d: Comparison at %d FAILED!\n", rank, i);
	    exit(-1);
	}
    }

}  /* end of check_buf() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

void
run_tests(unsigned int seed, char *snd, char *rcv, char *cmp)
{

    /* Do the easy ones first! */
    doit(seed, cmp, snd, rcv, 4);	check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv, 4 * 1024); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv, 40 * 1024); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv, 4 * 1024 * 1024); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv, 40 * 1024 * 1024); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);

    run1(seed, snd, rcv, cmp, 0);
    run1(seed, snd, rcv, cmp, 4 * 1024);
    run1(seed, snd, rcv, cmp, 40 * 1024);
    run1(seed, snd, rcv, cmp, 400 * 1024);
    run1(seed, snd, rcv, cmp, 4 * 1024 * 1024);
    run1(seed, snd, rcv, cmp, 40 * 1024 * 1024);
    run1(seed, snd, rcv, cmp, 13);
    run1(seed, snd, rcv, cmp, 83617); /* Some prime */

}  /* end of run_tests() */


void
run1(unsigned int seed, char *snd, char *rcv, char *cmp, int len)
{

    /* vary send side */
    doit(seed, cmp, snd, rcv, len + 1);	check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv, len + 2);	check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv, len + 3);	check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv, len + 5);	check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv, len + 6);	check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv, len + 7);	check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 1, rcv, len + 3); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv, len + 2); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 1, rcv, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 1, rcv, len + 4); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv, len + 4); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv, len + 4); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 1, rcv, len + 7); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv, len + 6); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 1, rcv, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 1, rcv, len + 8); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv, len + 8); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv, len + 8); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);

    /* vary receive side */
    doit(seed, cmp, snd, rcv + 1, len + 3); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 2, len + 2); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 3, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 1, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 2, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 3, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 1, len + 4); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 2, len + 4); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 3, len + 4); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 1, len + 7); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 2, len + 6); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 3, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 1, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 2, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 3, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 1, len + 8); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 2, len + 8); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd, rcv + 3, len + 8); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);


    /* vary send and receive side */
    doit(seed, cmp, snd + 1, rcv + 1, len + 3); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv + 2, len + 2); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv + 3, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 1, rcv + 1, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv + 2, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv + 3, len + 1); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 1, rcv + 1, len + 4); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv + 2, len + 4); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv + 3, len + 4); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 1, rcv + 1, len + 7); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv + 2, len + 6); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv + 3, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 1, rcv + 1, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv + 2, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv + 3, len + 5); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 1, rcv + 1, len + 8); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 2, rcv + 2, len + 8); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);
    doit(seed, cmp, snd + 3, rcv + 3, len + 8); check_buf(seed + R_SEED, rcv, BLOCKS * EightMB);

}  /* end of run1() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

void
doit(unsigned int seed, char *cmp, char *snd, char *rcv, int len)
{

int rank;
int i;
MPI_Request rreq;
MPI_Request sreq;
MPI_Status status;


    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    if (debug)   {
	printf("    %2d: doit(seed %d, snd %p, rcv %p, len %d)\n", rank, seed, snd, rcv, len);
    }

    if (rank == NODE0)   {
	printf("    snd %p, rcv %p, len %d)\n", snd, rcv, len);
    }

    if (rank == NODE0)   {
	MPI_Barrier(MPI_COMM_WORLD);

	/* Send the data */
	MPI_Isend(snd, len, MPI_CHAR, NODE1, TAG, MPI_COMM_WORLD, &sreq);
	MPI_Wait(&sreq, &status);
	if (debug)   {
	    printf("    %2d: Sent %d bytes\n", rank, len);
	}

    } else if (rank == NODE1)   {
	/* Save the region we are going to overwrite */
	memcpy(cmp, rcv, len);
	#ifdef DEBUG
	    printf("    %2d: Before ", rank);
	    for (i = 0; i < len; i++) {
		printf(" %4d", rcv[i]);
	    }
	    printf("\n");
	#endif /* DEBUG */

	/* Setup the receive buffer */
        MPI_Irecv(rcv, len, MPI_CHAR, NODE0, TAG, MPI_COMM_WORLD, &rreq);

	MPI_Barrier(MPI_COMM_WORLD);

	/* Wait for the data */
	MPI_Wait(&rreq, &status);
	/* This is mpi lite's version of count */
	if (status.st_length != len)  {
	    fprintf(stderr, "    %2d: FAILED: Expected %d bytes, but got %d\n", rank, len, status.st_length);
	    exit(-1);
	}

	if (debug)   {
	    printf("    %2d: Received %d bytes into %p\n", rank, len, rcv);
	}

	#ifdef DEBUG
	    printf("    %2d: After  ", rank);
	    for (i = 0; i < len; i++) {
		printf(" %4d", rcv[i]);
	    }
	    printf("\n");
	#endif /* DEBUG */

	/* Check the data */
	for (i = 0; i < len; i++) {
	    if (rcv[i] != snd[i])   {
		fprintf(stderr, "    %2d: Comparison at %d FAILED! %02x != %02x\n",
		    rank, i, rcv[i], cmp[i]);
		exit(-1);
	    }
	}

	/* Restore the receive buffer */
	memcpy(rcv, cmp, len);

    } else   {
	/* do nothing */
    }

}  /* end of doit() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

