// Copyright 2009-2010 Sandia Corporation. Under the terms
// of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
// Government retains certain rights in this software.
// 
// Copyright (c) 2009-2010, Sandia Corporation
// All rights reserved.
// 
// This file is part of the SST software package. For license
// information, see the LICENSE file in the top level directory of the
// distribution.


/*
** $Id:$
** Send random data and fixed patterns to other nodes and checksum or
** fully check everything. On fault, show what is wrong.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <malloc/malloc.h>	// For malloc_printf()
#include "mpi.h"


/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */
/*
** Functions
*/
int crc_check(int *bufr, int msg_size, int nnodes, int rank);
int full_check(int *bufr, int msg_size, int nnodes, int rank);
void fill_buffer(int *bufs, int msg_size);
void disp_progress(int n, int nnodes, int msg_size, time_t start, int *bufs);
int get_value(int seed);


/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */
/*
** Constants
*/
#define CHKSUM_SEED     (0xAAAA5555)
#define ITERATIONS      (10)
#define MSG_SIZE        (100 / sizeof(int))
#define MAX_NNODES	(4)


char store_bufs[MSG_SIZE * sizeof (int)];
char store_bufr[MAX_NNODES * MSG_SIZE * sizeof (int)];
char store_rreq[MAX_NNODES * sizeof(MPI_Request)];
char store_sreq[MAX_NNODES * sizeof(MPI_Request)];
char store_status[MAX_NNODES * sizeof(MPI_Status)];

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */


int
main(int argc, char **argv)
{

int i, n;
int *bufs;
int *bufr;
int *bufptr;
MPI_Request *sreq;
MPI_Request *rreq;
MPI_Status *status;
int msg_size, ntest;
time_t start_time;
unsigned int seed;
unsigned int chksum;
int do_check;
int rank, nnodes;


    malloc_printf("Hello 0\n");
    MPI_Init (&argc, &argv);
    MPI_Comm_size (MPI_COMM_WORLD, &nnodes);
    MPI_Comm_rank (MPI_COMM_WORLD, &rank);
    malloc_printf("Hello 1\n");
    malloc_printf("Hello 2 from node %d\n", rank);

    if (nnodes > MAX_NNODES)   {
	fprintf(stderr, "Too many nodes\n");
	exit(-1);
    }

    do_check= 1;

    msg_size= MSG_SIZE;
    ntest= ITERATIONS;

    if (rank == 0)   {
        printf("\n\nData integrety test\n");
        printf("===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++"
            "=++=++=++=++===\n");
        printf("    There will be %d iterations\n", ntest);
        printf("    In each iteration each node will send %d integers (%ld "
            "bytes each)\n", msg_size, sizeof(int));
        printf("    to each of the %d nodes\n", nnodes);
        if (do_check)   {
            printf("    Doing a full check on the data (not just CRC)\n");
        } else   {
            printf("    Doing a 32-bit CRC check on the data\n");
        }
        printf("    'a' through 'o' are specific patterns. '.' and ':' are "
            "random patterns\n");
    }

    bufs= (int *)store_bufs;
    bufr= (int *)store_bufr;
    rreq= (MPI_Request *)store_rreq;
    sreq= (MPI_Request *)store_sreq;
    status= (MPI_Status *)store_status;


    /*
    ** Fill the send buffer.
    ** The first word is the seed to srandom().
    ** The second word is the checksum of all the data words.
    ** The third through the last words are data.
    */
    seed= (int)time(NULL) * (rank + 117);
    if ((seed >= 0) && (seed <= 15))   {
        /* Seed is one of the fixed patterns; don't use it */
        seed= seed + 16;
    }
    srandom(seed);
    chksum= CHKSUM_SEED;
    for (i = 2; i < msg_size; i++) {
        bufs[i] = (int)random();
        chksum= chksum ^ (unsigned int)bufs[i];
    }
    bufs[0]= seed;
    bufs[1]= chksum;

    /*
    ** Pre-post all receives
    */
    bufr= memset(bufr, 0xd6, nnodes * msg_size * sizeof (int));
    for (i= 0; i < nnodes; i++)   {
        bufptr= bufr + (i * msg_size);
        MPI_Irecv(bufptr, msg_size, MPI_INT, i, 1001, MPI_COMM_WORLD, &rreq[i]);
    }

    if (rank == 0)   {
        printf("\nElapsed    Iterations        Aggregate\n");
        printf("Time                         data sent\n");
    }
    start_time= time(NULL);


    for (n = 0; n < ntest; n++) {
malloc_printf("Still OK!\n");
#if 0
        if (rank == 0)   {
            disp_progress(n, nnodes, msg_size, start_time, bufs);
        }

	MPI_Barrier(MPI_COMM_WORLD);

        /* send to everybody */
        for (i= 0; i < nnodes; i++)   {
            to= (rank + i + n) % nnodes;
            MPI_Isend(bufs, msg_size, MPI_INT, to, 1001, MPI_COMM_WORLD,
                &sreq[i]);
        }

        /* Wait for the receives to complete */
        MPI_Waitall(nnodes, rreq, status);


        /* Check the received data */
        if (!do_check)   {
            if (crc_check(bufr, msg_size, nnodes, rank) != 0)   {
                fprintf(stderr, "CRC check failed\n");
                exit(-1);
            }
        } else   {
            /* Fully Check the data */
            if (full_check(bufr, msg_size, nnodes, rank) != 0)   {
                fprintf(stderr, "Full check failed\n");
                exit(-1);
            }
        }


        /* Pre-post the receives */
        bufr= memset(bufr, 0xd6, nnodes * msg_size * sizeof (int));
        for (i= 0; i < nnodes; i++)   {
            bufptr= bufr + (i * msg_size);
            MPI_Irecv(bufptr, msg_size, MPI_INT, i, 1001, MPI_COMM_WORLD,
                &rreq[i]);
        }


        /* Make sure the sends have completed */
        MPI_Waitall(nnodes, sreq, &ignore);


        /* Change the data patterns around once in a while */
        if (((n + 1) % 50) == 0)   {
            fill_buffer(bufs, msg_size);
        }
#endif	/* 0 */
    }

    MPI_Barrier(MPI_COMM_WORLD);
    MPI_Finalize();

    exit(0);
}

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

int
get_value(int seed)
{

int value;


    switch (seed)   {
        case  0: value= 0x00000000; break;
        case  1: value= 0xffffffff; break;
        case  2: value= 0xaaaaaaaa; break;
        case  3: value= 0x55555555; break;
        case  4: value= 0x5a5a5a5a; break;
        case  5: value= 0xa5a5a5a5; break;
        case  6: value= 0xf0f0f0f0; break;
        case  7: value= 0x0f0f0f0f; break;
        case  8: value= 0xff00ff00; break;
        case  9: value= 0x00ff00ff; break;
        case 10: value= 0xffff0000; break;
        case 11: value= 0x0000ffff; break;
        case 12: value= 0xaa55aa55; break;
        case 13: value= 0x55aa55aa; break;
        case 14: value= 0xaaaa5555; break;
        case 15: value= 0x5555aaaa; break;
        default: value= (int)random();
    }

    return value;

}  /* end of get_value() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

int
crc_check(int *bufr, int msg_size, int nnodes, int rank)
{

int i, j;
int *bufptr;
int chksum;
int seed;
int error;
int abort;
int value;
int expected, gotten;
char err_line[1024];


    abort= 0;
    gotten= 0;
    expected= 0;
    for (i= 0; i < nnodes; i++)   {
        bufptr= bufr + (i * msg_size);
        chksum= CHKSUM_SEED;
        for (j = 2; j < msg_size; j++) {
            chksum= chksum ^ (unsigned int)bufptr[j];
        }
        if (chksum != bufptr[1])   {
            seed= bufptr[0];
            fprintf(stderr, "\n*** Checksum error on node %d from %d using "
                "pattern %d\n", rank, i, seed);

            srandom(seed);
            error= 0;
            for (j = 2; j < msg_size; j++) {
                value= get_value(seed);
                if (bufptr[j] != value)   {
                    if ((error == 0) && (j > 2))   {
                        /* Print last good compare as the lead in sequence */
                        fprintf(stderr, "Node %-4d   Word %6d: wanted 0x%08x   "
                            "got 0x%08x\n", rank, j - 1, expected, gotten);
                    }

                    sprintf(err_line, "Node %-4d   Word %6d: wanted 0x%08x   "
                        "got 0x%08x\n", rank, j, value, bufptr[j]);
                    if (error < 10)   {
                        fprintf(stderr, "%s", err_line);
                    }
                    error++;
                    abort= 1;
                } else   {
                    expected= value;
                    gotten= bufptr[j];
                    if (error)   {
                        if (error > 10)   {
                            fprintf(stderr, "Node %-4d   ...\n", rank);
                            fprintf(stderr, "%s", err_line);
                        }
                        fprintf(stderr, "Node %-4d   Word %6d: wanted 0x%08x   "
                            "got 0x%08x\n", rank, j, expected, gotten);
                        fprintf(stderr, "Node %-4d Data corruption in %d "
                            "words\n", rank, error);
                    }
                    error= 0;
                }
            }
            if (error)   {
                fprintf(stderr, "Node %-4d   ...\n", rank);
                fprintf(stderr, "%s", err_line);
                fprintf(stderr, "Node %-4d Data corruption in %d words\n",
                    rank, error);
            }
        }
    }

    if (abort)   {
        return -1;
    } else   {
        return 0;
    }

}  /* end of crc_check() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

int
full_check(int *bufr, int msg_size, int nnodes, int rank)
{

int i, j;
int *bufptr;
int chksum;
int seed;
int error;
int abort;
int value;
int expected, gotten;
char err_line[1024];


    abort= 0;
    gotten= 0;
    expected= 0;
    for (i= 0; i < nnodes; i++)   {
        bufptr= bufr + (i * msg_size);
        chksum= CHKSUM_SEED;
        seed= bufptr[0];
        srandom(seed);
        error= 0;
        for (j = 2; j < msg_size; j++) {
            value= get_value(seed);
            if (bufptr[j] != value)   {
                if ((error == 0) && (j > 2))   {
                    /* Print the last good compare as the lead in sequence */
                    fprintf(stderr, "Node %-4d   Word %6d: wanted 0x%08x   "
                        "got 0x%08x\n", rank, j - 1, expected, gotten);
                }

                sprintf(err_line, "Node %-4d   Word %6d: wanted 0x%08x   got "
                    "0x%08x\n", rank, j, value, bufptr[j]);
                if (error < 10)   {
                    /* Print a few of the wrong data pairs */
                    fprintf(stderr, "%s", err_line);
                }
                error++;
                abort= 1;
            } else   {
                expected= value;
                gotten= bufptr[j];
                if (error)   {
                    if (error > 10)   {
                        fprintf(stderr, "Node %-4d   ...\n", rank);
                        fprintf(stderr, "%s", err_line);
                    }
                    fprintf(stderr, "Node %-4d   Word %6d: wanted 0x%08x   "
                        "got 0x%08x\n", rank, j, expected, gotten);
                    fprintf(stderr, "Node %-4d Data corruption in %d words\n",
                        rank, error);
                    fprintf(stderr, "*** Compare error on node %d from %d "
                        "using pattern %d\n", rank, i, seed);
                }
                error= 0;
            }
            chksum= chksum ^ (unsigned int)bufptr[j];
        }

        if (error)   {
            fprintf(stderr, "Node %-4d   ...\n", rank);
            fprintf(stderr, "%s", err_line);
            fprintf(stderr, "Node %-4d Data corruption in %d words\n",
                rank, error);
            fprintf(stderr, "*** Compare error on node %d from %d using "
                "pattern %d\n", rank, i, seed);
            error= 0;
        }
        if (chksum != bufptr[1])   {
            fprintf(stderr, "Node %-4d Check sum error\n", rank);
            abort= 1;
        }
    }

    if (abort)   {
        return -1;
    } else   {
        return 0;
    }

}  /* end of full_check() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

void
fill_buffer(int *bufs, int msg_size)
{

int seed;
int chksum;
int i;
int value;


    seed= (int)time(NULL) ^ bufs[0];
    if ((seed >= 0) && (seed <= 15))   {
        /* Seed is one of the fixed patterns; don't use it */
        seed= seed + 16;
    }

    chksum= CHKSUM_SEED;
    if ((seed & 0x00000010) == 0x10)   {
        /* Use one of the fixed patterns */
        seed= seed & 0x0000000f;
        value= get_value(seed);
        for (i = 2; i < msg_size; i++) {
            bufs[i] = value;
            chksum= chksum ^ (unsigned int)bufs[i];
        }
    } else   {
        /* Use a random pattern */
        srandom(seed);
        for (i = 2; i < msg_size; i++) {
            bufs[i] = (int)random();
            chksum= chksum ^ (unsigned int)bufs[i];
        }
    }

    bufs[0]= seed;
    bufs[1]= chksum;

}  /* end of fill_buffer() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

void
disp_progress(int n, int nnodes, int msg_size, time_t start, int *bufs)
{

long long total;
double dtotal;
char *unit;
time_t elapsed;
int ss, mm, hh;
static int old_seed= 0;
static int flip= 0;


    if ((n % 250) == 0)   {
        /*
        ** In n iterations
        ** nnodes send a message to nnodes
        ** Each message is msg_size integers
        */
        total= (long long)nnodes * (long long)n * (long long)msg_size *
                    (long long)sizeof(int) * (long long)(nnodes);
#ifdef PETABYTES
        if (total >= 1000000000000000)   {
            unit= "Peta Bytes";
            dtotal= total / 1000000000000000.0;
        } else
#endif /* PETABYTES */
#ifdef TERABYTES
	if (total >= 1000000000000)   {
            unit= "Tera Bytes";
            dtotal= total / 1000000000000.0;
        } else
#endif /* TERABYTES */
	if (total >= 1000000000)   {
            unit= "Giga Bytes";
            dtotal= total / 1000000000.0;
        } else if (total >= 1000000)   {
            unit= "Mega Bytes";
            dtotal= total / 1000000.0;
        } else if (total >= 1000)   {
            unit= "Kilo Bytes";
            dtotal= total / 1000.0;
        } else   {
            unit= "Bytes";
            dtotal= total;
        }

        elapsed= time(NULL) - start;
        hh= elapsed / (60 * 60);
        elapsed= elapsed - (hh * 60 * 60);
        mm= elapsed / 60;
        elapsed= elapsed - (mm * 60);
        ss= elapsed;
        printf("\n%3d:%02d:%02d  %-8d     %14.3f %-11s ", hh, mm, ss, n,
            dtotal, unit);
        fflush(stdout);

    } else if ((n % 10) == 0)   {

        /*
        ** Print a letter indicating whether we are sending random data
        ** or one of the predefined patters.
        */
        if ((bufs[0] >= 0) && (bufs[0] <= 15))   {
            printf("%c", bufs[0] + 'a');
        } else   {
            if (bufs[0] != old_seed)   {
                old_seed= bufs[0];
                flip= (flip ^ 0x1) & 0x1;
            }
            if (flip)   {
                printf(".");
            } else   {
                printf(":");
            }
        }
        fflush(stdout);
    }

}  /* end of disp_progress() */

/* ===++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=++=== */

